# sage_setup: distribution = sagemath-graphs
# sage.doctest: needs sage.graphs
r"""
Quiver mutation types

AUTHORS:

- Gregg Musiker (2012, initial version)
- Christian Stump (2012, initial version)
- Hugh Thomas (2012, initial version)
"""
# ****************************************************************************
#       Copyright (C) 2011 Gregg Musiker <gmusiker@gmail.com>
#                          Christian Stump <christian.stump@gmail.com>
#                          Hugh Thomas <hugh@math.unb.ca>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  https://www.gnu.org/licenses/
# ***************************************************************************
from __future__ import annotations
from pathlib import Path
import pickle

from sage.structure.sage_object import SageObject
from copy import copy
from sage.structure.unique_representation import UniqueRepresentation
from sage.misc.cachefunc import cached_method
from sage.misc.lazy_import import lazy_import
from sage.rings.integer_ring import ZZ
from sage.rings.infinity import infinity
from sage.graphs.digraph import DiGraph
from sage.graphs.graph import Graph
from sage.arith.misc import binomial, euler_phi
from sage.misc.misc_c import prod

lazy_import('sage.matrix.constructor', 'matrix')


class QuiverMutationTypeFactory(SageObject):
    def __call__(self, *args):
        """
        For a detailed description, see :meth:`QuiverMutationType`.

        EXAMPLES::

            sage: from sage.combinat.cluster_algebra_quiver.quiver_mutation_type import QuiverMutationTypeFactory
            sage: QuiverMutationTypeFactory()
            QuiverMutationType
        """
        # get data as arguments or as list/tuple
        if len(args) == 1:
            data = args[0]
        else:
            data = args

        # data is a QuiverMutationType
        if isinstance(data, QuiverMutationType_Irreducible):
            return data
        elif isinstance(data, QuiverMutationType_Reducible):
            return data

        # check that data is a tuple or list
        if isinstance(data, tuple) and data:
            pass
        elif isinstance(data, list) and data:
            data = tuple(data)
        else:
            _mutation_type_error(data)

        # check for reducible types
        if all(isinstance(data_component, (list, tuple,
                                           QuiverMutationType_Irreducible))
               for data_component in data):
            if len(data) == 1:
                return QuiverMutationType(data[0])
            else:
                data = tuple(QuiverMutationType(comp) for comp in data)
                return QuiverMutationType_Reducible(*data)

        # check for irreducible types
        if len(data) == 2:
            data = (data[0], data[1], None)
        elif len(data) == 3:
            pass
        else:
            _mutation_type_error(data)

        if isinstance(data[2], list):
            data = (data[0], data[1], tuple(data[2]))
        if isinstance(data[1], list):
            data = (data[0], tuple(data[1]), data[2])

        # mutation type casting
        if data == ('D', 2, None):
            return QuiverMutationType(('A', 1, None), ('A', 1, None))
        elif data == ('D', 3, None):
            data = ('A', 3, None)
        elif data == ('C', 2, None):
            data = ('B', 2, None)
        elif data == ('E', 9, None):
            data = ('E', 8, 1)
        elif data[0] == 'A' and data[2] == 1 and isinstance(data[1], tuple) and len(data[1]) == 2 and min(data[1]) == 0:
            if max(data[1]) == 0:
                pass
            elif max(data[1]) == 1:
                data = ('A', 1, None)
            elif max(data[1]) == 2:
                return QuiverMutationType(('A', 1, None), ('A', 1, None))
            elif max(data[1]) == 3:
                data = ('A', 3, None)
            else:
                data = ('D', max(data[1]), None)
        elif data[0] == 'GR' and data[2] is None and isinstance(data[1], tuple) and len(data[1]) == 2 and data[1][1] > data[1][0]:
            if min(data[1]) > max(data[1]) / 2 and max(data[1]) != min(data[1]) + 1:
                data = (data[0], (max(data[1]) - min(data[1]), max(data[1])), data[2])
            if min(data[1]) == 2 and max(data[1]) > 3:
                data = ('A', max(data[1]) - 3, None)
            elif data[1] == (3, 6):
                data = ('D', 4, None)
            elif data[1] == (3, 7):
                data = ('E', 6, None)
            elif data[1] == (3, 8):
                data = ('E', 8, None)
            elif data[1] == (3, 9):
                data = ('E', 8, [1, 1])
            elif data[1] == (4, 8):
                data = ('E', 7, [1, 1])
        elif data == ('TR', 1, None):
            data = ('A', 1, None)
        elif data == ('TR', 2, None):
            data = ('A', 3, None)
        elif data == ('TR', 3, None):
            data = ('D', 6, None)
        elif data == ('TR', 4, None):
            data = ('E', 8, (1, 1))
        # mutation type casting from Kac conventions
        elif data == ('A', 1, 1):
            data = ('A', (1, 1), 1)
        elif data[0] == 'B' and data[2] == 1:
            if data[1] == 2:
                data = ('CC', 2, 1)
            elif data[1] > 2:
                data = ('BD', data[1], 1)
        elif data[0] == 'B' and data[2] == -1:
            if data[1] == 2:
                data = ('BB', 2, 1)
            elif data[1] > 2:
                data = ('CD', data[1], 1)
        elif data[0] == 'C' and data[1] > 1 and data[2] == 1:
            data = ('CC', data[1], 1)
        elif data[0] == 'C' and data[1] > 1 and data[2] == -1:
            data = ('BB', data[1], 1)
        elif data == ('A', 2, 2):
            data = ('BC', 1, 1)
        elif data[0] == 'A' and data[1] in ZZ and data[1] > 1 and data[1] % 2 == 0 and data[2] == 2:
            data = ('BC', data[1] // 2, 1)
        elif data[0] == 'A' and data[1] in ZZ and data[1] > 3 and data[1] % 2 and data[2] == 2:
            data = ('CD', (data[1] + 1) // 2, 1)
        # We think of ('A',3, 2) as ('D',3, 2)
        elif data == ('A', 3, 2):
            data = ('BB', 2, 1)
        elif data[0] == 'D' and data[1] in ZZ and data[1] > 2 and data[2] == 2:
            data = ('BB', data[1] - 1, 1)
        elif data == ('E', 6, 2):
            data = ('F', 4, -1)
        elif data == ('D', 4, 3):
            data = ('G', 2, -1)
        elif data == ('F', 4, (2, 1)):
            data = ('F', 4, (1, 2))
        elif data == ('G', 2, (3, 1)):
            data = ('G', 2, (1, 3))
        elif data[0] == 'T' and data[2] is None:
            data = (data[0], tuple(sorted(data[1])), data[2])
            r, p, q = data[1]
            if r == 1:
                data = ('A', p + q - 1, None)
            elif r == p == 2:
                data = ('D', q + 2, None)
            elif r == 2 and p == 3:
                if q in (3, 4, 5):
                    data = ('E', q + 3, None)
                elif q == 6:
                    data = ('E', 8, 1)
                else:
                    data = ('E', q + 3, None)
            elif r == 2 and p == q == 4:
                data = ('E', 7, 1)
            elif r == p == q == 3:
                data = ('E', 6, 1)
        elif data[0] == 'R2' and data[2] is None and all(data[1][i] in ZZ and data[1][i] > 0 for i in [0, 1]):
            data = (data[0], tuple(sorted(data[1])), data[2])
            if data[1] == (1, 1):
                data = ('A', 2, None)
            elif data[1] == (1, 2):
                data = ('B', 2, None)
            elif data[1] == (1, 3):
                data = ('G', 2, None)
            elif data[1] == (1, 4):
                data = ('BC', 1, 1)
            elif data[1] == (2, 2):
                data = ('A', (1, 1), 1)

        # setting the parameters and returning the mutation type
        letter, rank, twist = data
        if not isinstance(letter, str):
            _mutation_type_error(data)
        if isinstance(rank, list):
            rank = tuple(rank)
        if isinstance(twist, list):
            twist = tuple(twist)
        return QuiverMutationType_Irreducible(letter, rank, twist)

    def _repr_(self) -> str:
        """
        Return the string representation of ``self``.

        EXAMPLES::

            sage: QuiverMutationType  # indirect doctest
            QuiverMutationType
        """
        return "QuiverMutationType"

    def samples(self, finite=None, affine=None, elliptic=None,
                mutation_finite=None):
        """
        Return a sample of the available quiver mutations types.

        INPUT:

        - ``finite``

        - ``affine``

        - ``elliptic``

        - ``mutation_finite``

        All four input keywords default values are ``None``. If
        set to ``True`` or ``False``, only these samples are returned.

        EXAMPLES::

            sage: QuiverMutationType.samples()
            [['A', 1], ['A', 5], ['B', 2], ['B', 5], ['C', 3],
             ['C', 5], [ ['A', 1], ['A', 1] ], ['D', 5], ['E', 6],
             ['E', 7], ['E', 8], ['F', 4], ['G', 2],
             ['A', [1, 1], 1], ['A', [4, 5], 1], ['D', 4, 1],
             ['BB', 5, 1], ['E', 6, [1, 1]], ['E', 7, [1, 1]],
             ['R2', [1, 5]], ['R2', [3, 5]], ['E', 10], ['BE', 5],
             ['GR', [3, 10]], ['T', [3, 3, 4]]]

            sage: QuiverMutationType.samples(finite=True)
            [['A', 1], ['A', 5], ['B', 2], ['B', 5], ['C', 3],
             ['C', 5], [ ['A', 1], ['A', 1] ], ['D', 5], ['E', 6],
             ['E', 7], ['E', 8], ['F', 4], ['G', 2]]

            sage: QuiverMutationType.samples(affine=True)
            [['A', [1, 1], 1], ['A', [4, 5], 1], ['D', 4, 1], ['BB', 5, 1]]

            sage: QuiverMutationType.samples(elliptic=True)
            [['E', 6, [1, 1]], ['E', 7, [1, 1]]]

            sage: QuiverMutationType.samples(mutation_finite=False)
            [['R2', [1, 5]], ['R2', [3, 5]], ['E', 10], ['BE', 5],
             ['GR', [3, 10]], ['T', [3, 3, 4]]]
        """
        result = self._samples()
        if finite is not None:
            result = [t for t in result if t.is_finite() == finite]
        if affine is not None:
            result = [t for t in result if t.is_affine() == affine]
        if elliptic is not None:
            result = [t for t in result if t.is_elliptic() == elliptic]
        if mutation_finite is not None:
            result = [t for t in result
                      if t.is_mutation_finite() == mutation_finite]
        return result

    @cached_method
    def _samples(self):
        """
        Return a list of sample of available Cartan types.

        EXAMPLES::

            sage: X = QuiverMutationType._samples()
        """
        finite_types = \
            [QuiverMutationType(t) for t in [['A', 1], ['A', 5], ['B', 2], ['B', 5],
                                             ['C', 3], ['C', 5], ['D', 2], ['D', 5],
                                             ["E", 6], ["E", 7], ["E", 8], ["F", 4],
                                             ["G", 2]]]
        affine_types = \
            [QuiverMutationType(t) for t in [['A', [1, 1], 1], ['A', [4, 5], 1], ['D', 4, 1], ['BB', 5, 1]]]
        elliptic_types = \
            [QuiverMutationType(t) for t in [['E', 6, [1, 1]], ['E', 7, [1, 1]]]]
        mutation_finite_types = \
            [QuiverMutationType(t) for t in [['R2', (1, 5)], ['R2', (3, 5)]]]
        mutation_infinite_types = \
            [QuiverMutationType(t) for t in [['E', 10], ['BE', 5], ['GR', (3, 10)], ['T', (3, 3, 4)]]]

        return finite_types + affine_types + elliptic_types + mutation_finite_types + mutation_infinite_types


QuiverMutationType = QuiverMutationTypeFactory()


QuiverMutationType.__doc__ = \
    r"""

*Quiver mutation types* can be seen as a slight generalization of
 *generalized Cartan types*.

Background on generalized Cartan types can be found at

        :wikipedia:`Generalized_Cartan_matrix`

For the compendium on the cluster algebra and quiver package in Sage see [MS2011]_

A `B`-matrix is a skew-symmetrizable `( n \times n )`-matrix `M`.
I.e., there exists an invertible diagonal matrix `D` such that `DM` is
skew-symmetric.  `M` can be encoded as a *quiver* by having a directed
edge from vertex `i` to vertex `j` with label `(a,b)` if `a = M_{i,j}
> 0` and `b = M_{j,i} < 0`.  We consider quivers up to *mutation
equivalence*.

To a quiver mutation type we can associate a *generalized Cartan type*
by sending `M` to the generalized Cartan matrix `C(M)` obtained by
replacing all positive entries by their negatives and adding `2`'s on
the main diagonal.

:class:`QuiverMutationType` constructs a quiver mutation type object. For
more detail on the possible different types, please see the
compendium.

INPUT:

The input consists either of a quiver mutation type, or of a
``letter`` (a string), a ``rank`` (one integer or a list/tuple of
integers), and an optional ``twist`` (an integer or a list of
integers).  There are several different naming conventions for quiver
mutation types.

- Finite type -- ``letter`` is a Dynkin type (A-G), and ``rank`` is
  the rank.

- Affine type -- there is more than one convention for naming affine
  types.

    * Kac's notation: ``letter`` is a Dynkin type, ``rank`` is the
      rank of the associated finite Dynkin diagram, and ``twist`` is the
      twist, which could be 1, 2, or 3.  In the special case of affine
      type A, there is more than one quiver mutation type associated to
      the Cartan type.  In this case only, ``rank`` is a pair of integers
      (i,j), giving the number of edges pointing clockwise and the number
      of edges pointing counter-clockwise.  The total number of vertices
      is given by i+j in this case.

    * Naive notation: ``letter`` is one of 'BB', 'BC', 'BD', 'CC',
      'CD'.  The name specifies the two ends of the diagram, which are
      joined by a path.  The total number of vertices is given by
      ``rank +1`` (to match the indexing people expect because these
      are affine types).  In general, ``rank`` must be large enough
      for the picture to make sense, but we accept ``letter`` is
      ``BC`` and ``rank=1``.

    * Macdonald notation: for the dual of an untwisted affine type
      (such as ['C', 6, 1]), we accept a twist of -1 (i.e.,
      ['C',6,-1]).

- Elliptic type -- ``letter`` is a Dynkin type, ``rank`` is the rank
  of the finite Dynkin diagram, and ``twist`` is a tuple of two
  integers.  We follow Saito's notation.

- Other shapes:

    * Rank 2: ``letter`` is 'R2', and ``rank`` is a pair of integers
      specifying the label on the unique edge.

    * Triangle: ``letter`` is ``TR``, and ``rank`` is the number of
      vertices along a side.

    * T: This defines a quiver shaped like a T.  ``letter`` is 'T',
      and the ``rank`` is a triple, whose entries specify the number
      of vertices along each path from the branch point (counting the
      branch point).

    * Grassmannian: This defines the cluster algebra (without
      coefficients) corresponding to the cluster algebra with
      coefficients which is the coordinate ring of a Grassmannian.
      ``letter`` is 'GR'.  ``rank`` is a pair of integers (`k`, `n`)
      with 'k' < 'n' specifying the Grassmannian of `k`-planes in
      `n`-space.  This defines a quiver given by a (k-1) x (n-k-1)
      grid where each square is cyclically oriented.

    * Exceptional mutation finite quivers: The two exceptional
      mutation finite quivers, found by Derksen-Owen, have ``letter``
      as 'X' and ``rank`` 6 or 7, equal to the number of vertices.

    * AE, BE, CE, DE: Quivers are built of one end which looks like
      type (affine A), B, C, or D, and the other end which looks like
      type E (i.e., it consists of two antennae, one of length one,
      and one of length two).  ``letter`` is 'AE', 'BE', 'CE', or
      'DE', and ``rank`` is the total number of vertices.  Note that
      'AE' is of a slightly different form and requires ``rank`` to be
      a pair of integers (i,j) just as in the case of affine type A.
      See Exercise 4.3 in Kac's book Infinite Dimensional Lie Algebras
      for more details.

    * Infinite type E: It is also possible to obtain infinite-type E
      quivers by specifying ``letter`` as 'E' and ``rank`` as the
      number of vertices.

REFERENCES:

- A good reference for finite and affine Dynkin diagrams, including
  Kac's notation, is the :wikipedia:`Dynkin_diagram`.

- A good reference for the skew-symmetrizable elliptic diagrams is
  "Cluster algebras of finite mutation type via unfolding" by
  A. Felikson, M. Shapiro, and P. Tumarkin, [FST2012]_.

EXAMPLES:

Finite types::

    sage: QuiverMutationType('A', 1)
    ['A', 1]
    sage: QuiverMutationType('A',5)
    ['A', 5]

    sage: QuiverMutationType('B', 2)
    ['B', 2]
    sage: QuiverMutationType('B',5)
    ['B', 5]

    sage: QuiverMutationType('C', 2)
    ['B', 2]
    sage: QuiverMutationType('C',5)
    ['C', 5]

    sage: QuiverMutationType('D', 2)
    [ ['A', 1], ['A', 1] ]
    sage: QuiverMutationType('D',3)
    ['A', 3]
    sage: QuiverMutationType('D',4)
    ['D', 4]

    sage: QuiverMutationType('E',6)
    ['E', 6]

    sage: QuiverMutationType('G', 2)
    ['G', 2]

    sage: QuiverMutationType('A',(1,0), 1)
    ['A', 1]

    sage: QuiverMutationType('A',(2,0), 1)
    [ ['A', 1], ['A', 1] ]

    sage: QuiverMutationType('A',(7,0), 1)
    ['D', 7]

Affine types::

    sage: QuiverMutationType('A',(1, 1), 1)
    ['A', [1, 1], 1]
    sage: QuiverMutationType('A',(2,4), 1)
    ['A', [2, 4], 1]

    sage: QuiverMutationType('BB', 2, 1)
    ['BB', 2, 1]
    sage: QuiverMutationType('BB',4, 1)
    ['BB', 4, 1]

    sage: QuiverMutationType('CC', 2, 1)
    ['CC', 2, 1]
    sage: QuiverMutationType('CC',4, 1)
    ['CC', 4, 1]

    sage: QuiverMutationType('BC', 1, 1)
    ['BC', 1, 1]
    sage: QuiverMutationType('BC',5, 1)
    ['BC', 5, 1]

    sage: QuiverMutationType('BD',3, 1)
    ['BD', 3, 1]
    sage: QuiverMutationType('BD',5, 1)
    ['BD', 5, 1]

    sage: QuiverMutationType('CD',3, 1)
    ['CD', 3, 1]
    sage: QuiverMutationType('CD',5, 1)
    ['CD', 5, 1]

    sage: QuiverMutationType('D',4, 1)
    ['D', 4, 1]
    sage: QuiverMutationType('D',6, 1)
    ['D', 6, 1]

    sage: QuiverMutationType('E',6, 1)
    ['E', 6, 1]
    sage: QuiverMutationType('E',7, 1)
    ['E', 7, 1]
    sage: QuiverMutationType('E',8, 1)
    ['E', 8, 1]

    sage: QuiverMutationType('F',4, 1)
    ['F', 4, 1]
    sage: QuiverMutationType('F',4,-1)
    ['F', 4, -1]

    sage: QuiverMutationType('G', 2, 1)
    ['G', 2, 1]
    sage: QuiverMutationType('G', 2,-1)
    ['G', 2, -1]
    sage: QuiverMutationType('A',3, 2) == QuiverMutationType('D',3, 2)
    True

Affine types using Kac's Notation::

    sage: QuiverMutationType('A', 1, 1)
    ['A', [1, 1], 1]
    sage: QuiverMutationType('B',5, 1)
    ['BD', 5, 1]
    sage: QuiverMutationType('C',5, 1)
    ['CC', 5, 1]
    sage: QuiverMutationType('A', 2, 2)
    ['BC', 1, 1]
    sage: QuiverMutationType('A',7, 2)
    ['CD', 4, 1]
    sage: QuiverMutationType('A',8, 2)
    ['BC', 4, 1]
    sage: QuiverMutationType('D',6, 2)
    ['BB', 5, 1]
    sage: QuiverMutationType('E',6, 2)
    ['F', 4, -1]
    sage: QuiverMutationType('D',4,3)
    ['G', 2, -1]

Elliptic types::

    sage: QuiverMutationType('E',6,[1, 1])
    ['E', 6, [1, 1]]
    sage: QuiverMutationType('F',4,[2, 1])
    ['F', 4, [1, 2]]
    sage: QuiverMutationType('G', 2,[3,3])
    ['G', 2, [3, 3]]

Mutation finite types:

Rank 2 cases::

    sage: QuiverMutationType('R2',(1, 1))
    ['A', 2]
    sage: QuiverMutationType('R2',(1, 2))
    ['B', 2]
    sage: QuiverMutationType('R2',(1,3))
    ['G', 2]
    sage: QuiverMutationType('R2',(1,4))
    ['BC', 1, 1]
    sage: QuiverMutationType('R2',(1,5))
    ['R2', [1, 5]]
    sage: QuiverMutationType('R2',(2, 2))
    ['A', [1, 1], 1]
    sage: QuiverMutationType('R2',(3,5))
    ['R2', [3, 5]]

Exceptional Derksen-Owen quivers::

    sage: QuiverMutationType('X',6)
    ['X', 6]


(Mainly) mutation infinite types:

Infinite type E::

    sage: QuiverMutationType('E',9)
    ['E', 8, 1]
    sage: QuiverMutationType('E', 10)
    ['E', 10]
    sage: QuiverMutationType('E', 12)
    ['E', 12]

    sage: QuiverMutationType('AE',(2,3))
    ['AE', [2, 3]]
    sage: QuiverMutationType('BE',5)
    ['BE', 5]
    sage: QuiverMutationType('CE',5)
    ['CE', 5]
    sage: QuiverMutationType('DE',6)
    ['DE', 6]

Grassmannian types::

    sage: QuiverMutationType('GR',(2,4))
    ['A', 1]
    sage: QuiverMutationType('GR',(2,6))
    ['A', 3]
    sage: QuiverMutationType('GR',(3,6))
    ['D', 4]
    sage: QuiverMutationType('GR',(3,7))
    ['E', 6]
    sage: QuiverMutationType('GR',(3,8))
    ['E', 8]
    sage: QuiverMutationType('GR',(3, 10))
    ['GR', [3, 10]]

Triangular types::

    sage: QuiverMutationType('TR', 2)
    ['A', 3]
    sage: QuiverMutationType('TR',3)
    ['D', 6]
    sage: QuiverMutationType('TR',4)
    ['E', 8, [1, 1]]
    sage: QuiverMutationType('TR',5)
    ['TR', 5]

T types::

    sage: QuiverMutationType('T',(1, 1, 1))
    ['A', 1]
    sage: QuiverMutationType('T',(1, 1,4))
    ['A', 4]
    sage: QuiverMutationType('T',(1,4,4))
    ['A', 7]
    sage: QuiverMutationType('T',(2, 2, 2))
    ['D', 4]
    sage: QuiverMutationType('T',(2, 2,4))
    ['D', 6]
    sage: QuiverMutationType('T',(2,3,3))
    ['E', 6]
    sage: QuiverMutationType('T',(2,3,4))
    ['E', 7]
    sage: QuiverMutationType('T',(2,3,5))
    ['E', 8]
    sage: QuiverMutationType('T',(2,3,6))
    ['E', 8, 1]
    sage: QuiverMutationType('T',(2,3,7))
    ['E', 10]
    sage: QuiverMutationType('T',(3,3,3))
    ['E', 6, 1]
    sage: QuiverMutationType('T',(3,3,4))
    ['T', [3, 3, 4]]

Reducible types::

    sage: QuiverMutationType(['A',3],['B',4])
    [ ['A', 3], ['B', 4] ]
"""


class QuiverMutationType_abstract(UniqueRepresentation, SageObject):
    """
    EXAMPLES::

        sage: mut_type1 = QuiverMutationType('A',5)
        sage: mut_type2 = QuiverMutationType('A',5)
        sage: mut_type3 = QuiverMutationType('A',6)
        sage: mut_type1 == mut_type2
        True
        sage: mut_type1 == mut_type3
        False
    """

    def _repr_(self):
        """
        Return the string representation of ``self``.

        EXAMPLES::

            sage: QuiverMutationType(['A', 2])  # indirect doctest
            ['A', 2]
        """
        return self._description

    def plot(self, circular=False, directed=True):
        """
        Return the plot of the underlying graph or digraph of ``self``.

        INPUT:

        - ``circular`` -- boolean (default: ``False``); if ``True``, the
          circular plot is chosen, otherwise >>spring<< is used

        - ``directed`` -- boolean (default: ``True``); if ``True``, the
          directed version is shown, otherwise the undirected

        EXAMPLES::

            sage: QMT = QuiverMutationType(['A',5])
            sage: pl = QMT.plot()                                                       # needs sage.plot sage.symbolic
            sage: pl = QMT.plot(circular=True)                                          # needs sage.plot sage.symbolic
        """
        return self.standard_quiver().plot(circular=circular, directed=directed)

    def show(self, circular=False, directed=True):
        """
        Show the plot of the underlying digraph of ``self``.

        INPUT:

        - ``circular`` -- boolean (default: ``False``); if ``True``, the
          circular plot is chosen, otherwise >>spring<< is used

        - ``directed`` -- boolean (default: ``True``); if ``True``, the
          directed version is shown, otherwise the undirected

        TESTS::

            sage: QMT = QuiverMutationType(['A', 5])
            sage: QMT.show()                    # long time                             # needs sage.plot sage.symbolic
        """
        self.plot(circular=circular, directed=directed).show()

    def letter(self):
        """
        Return the classification letter of ``self``.

        EXAMPLES::

            sage: mut_type = QuiverMutationType( ['A',5] ); mut_type
            ['A', 5]
            sage: mut_type.letter()
            'A'

            sage: mut_type = QuiverMutationType( ['BC',5, 1] ); mut_type
            ['BC', 5, 1]
            sage: mut_type.letter()
            'BC'

            sage: mut_type = QuiverMutationType(['A',3],['B',3]); mut_type
            [ ['A', 3], ['B', 3] ]
            sage: mut_type.letter()
            'A x B'

            sage: mut_type = QuiverMutationType(['A',3],['B',3],['X',6]); mut_type
            [ ['A', 3], ['B', 3], ['X', 6] ]
            sage: mut_type.letter()
            'A x B x X'
        """
        return self._letter

    def rank(self):
        """
        Return the rank in the standard quiver of ``self``.

        The rank is the number of vertices.

        EXAMPLES::

            sage: mut_type = QuiverMutationType( ['A',5] ); mut_type
            ['A', 5]
            sage: mut_type.rank()
            5

            sage: mut_type = QuiverMutationType( ['A',[4,5], 1] ); mut_type
            ['A', [4, 5], 1]
            sage: mut_type.rank()
            9

            sage: mut_type = QuiverMutationType( ['BC',5, 1] ); mut_type
            ['BC', 5, 1]
            sage: mut_type.rank()
            6

            sage: mut_type = QuiverMutationType(['A',3],['B',3]); mut_type
            [ ['A', 3], ['B', 3] ]
            sage: mut_type.rank()
            6

            sage: mut_type = QuiverMutationType(['A',3],['B',3],['X',6]); mut_type
            [ ['A', 3], ['B', 3], ['X', 6] ]
            sage: mut_type.rank()
            12
        """
        return self._rank

    @cached_method
    def b_matrix(self):
        """
        Return the B-matrix of the standard quiver of ``self``.

        The conventions for B-matrices agree with Fomin-Zelevinsky (up
        to a reordering of the simple roots).

        EXAMPLES::

            sage: mut_type = QuiverMutationType(['A',5]); mut_type
            ['A', 5]
            sage: mut_type.b_matrix()                                                   # needs sage.modules
            [ 0  1  0  0  0]
            [-1  0 -1  0  0]
            [ 0  1  0  1  0]
            [ 0  0 -1  0 -1]
            [ 0  0  0  1  0]

            sage: mut_type = QuiverMutationType(['A',3],['B',3]); mut_type
            [ ['A', 3], ['B', 3] ]
            sage: mut_type.b_matrix()                                                   # needs sage.modules
            [ 0  1  0  0  0  0]
            [-1  0 -1  0  0  0]
            [ 0  1  0  0  0  0]
            [ 0  0  0  0  1  0]
            [ 0  0  0 -1  0 -1]
            [ 0  0  0  0  2  0]
        """
        return _edge_list_to_matrix(self._digraph.edges(sort=True), list(range(self._rank)), [])

    @cached_method
    def standard_quiver(self):
        """
        Return the standard quiver of ``self``.

        EXAMPLES::

            sage: mut_type = QuiverMutationType( ['A',5] ); mut_type
            ['A', 5]
            sage: mut_type.standard_quiver()                                            # needs sage.modules
            Quiver on 5 vertices of type ['A', 5]

            sage: mut_type = QuiverMutationType( ['A',[5,3], 1] ); mut_type
            ['A', [3, 5], 1]
            sage: mut_type.standard_quiver()                                            # needs sage.modules
            Quiver on 8 vertices of type ['A', [3, 5], 1]

            sage: mut_type = QuiverMutationType(['A',3],['B',3]); mut_type
            [ ['A', 3], ['B', 3] ]
            sage: mut_type.standard_quiver()                                            # needs sage.modules
            Quiver on 6 vertices of type [ ['A', 3], ['B', 3] ]

            sage: mut_type = QuiverMutationType(['A',3],['B',3],['X',6]); mut_type
            [ ['A', 3], ['B', 3], ['X', 6] ]
            sage: mut_type.standard_quiver()                                            # needs sage.modules
            Quiver on 12 vertices of type [ ['A', 3], ['B', 3], ['X', 6] ]
        """
        from .quiver import ClusterQuiver
        Q = ClusterQuiver(self._digraph)
        Q._mutation_type = self
        return Q

    @cached_method
    def cartan_matrix(self):
        """
        Return the Cartan matrix of ``self``.

        Note that (up to a reordering of the simple roots) the convention for
        the definition of Cartan matrix, used here and elsewhere in Sage,
        agrees with the conventions of Kac, Fulton-Harris, and
        Fomin-Zelevinsky, but disagrees with the convention of Bourbaki.
        The `(i,j)` entry is `2(\\alpha_i,\\alpha_j)/(\\alpha_i,\\alpha_i)`.

        EXAMPLES::

            sage: mut_type = QuiverMutationType(['A',5]); mut_type
            ['A', 5]
            sage: mut_type.cartan_matrix()                                              # needs sage.modules
            [ 2 -1  0  0  0]
            [-1  2 -1  0  0]
            [ 0 -1  2 -1  0]
            [ 0  0 -1  2 -1]
            [ 0  0  0 -1  2]

            sage: mut_type = QuiverMutationType(['A',3],['B',3]); mut_type
            [ ['A', 3], ['B', 3] ]
            sage: mut_type.cartan_matrix()                                              # needs sage.modules
            [ 2 -1  0  0  0  0]
            [-1  2 -1  0  0  0]
            [ 0 -1  2  0  0  0]
            [ 0  0  0  2 -1  0]
            [ 0  0  0 -1  2 -1]
            [ 0  0  0  0 -2  2]
        """
        # as soon as CartanMatrix is implemented we should use it here:
        # from sage.combinat.root_system.cartan_matrix import CartanMatrix
        cmat = copy(self.b_matrix())
        for i, j in cmat.nonzero_positions():
            a = cmat[i, j]
            if a > 0:
                cmat[i, j] = -a
        for i in range(self._rank):
            cmat[i, i] = 2
        # return CartanMatrix(cmat)
        return cmat

    def is_irreducible(self):
        """
        Return ``True`` if ``self`` is irreducible.

        EXAMPLES::

            sage: mt = QuiverMutationType(['A', 2])
            sage: mt.is_irreducible()
            True
        """
        return self._info['irreducible']

    def is_mutation_finite(self):
        """
        Return ``True`` if ``self`` is of finite mutation type.

        This means that its mutation class has only finitely many
        different B-matrices.

        EXAMPLES::

            sage: mt = QuiverMutationType(['D',5, 1])
            sage: mt.is_mutation_finite()
            True
        """
        return self._info['mutation_finite']

    def is_simply_laced(self):
        """
        Return ``True`` if ``self`` is simply laced.

        This means that the only arrows that appear in the quiver of
        ``self`` are single unlabelled arrows.

        EXAMPLES::

            sage: mt = QuiverMutationType(['A', 2])
            sage: mt.is_simply_laced()
            True

            sage: mt = QuiverMutationType(['B', 2])
            sage: mt.is_simply_laced()
            False

            sage: mt = QuiverMutationType(['A',(1, 1), 1])
            sage: mt.is_simply_laced()
            False
        """
        return self._info['simply_laced']

    def is_skew_symmetric(self):
        """
        Return ``True`` if the B-matrix of ``self`` is skew-symmetric.

        EXAMPLES::

            sage: mt = QuiverMutationType(['A', 2])
            sage: mt.is_skew_symmetric()
            True

            sage: mt = QuiverMutationType(['B', 2])
            sage: mt.is_skew_symmetric()
            False

            sage: mt = QuiverMutationType(['A',(1, 1), 1])
            sage: mt.is_skew_symmetric()
            True
        """
        return self._info['skew_symmetric']

    def is_finite(self):
        """
        Return ``True`` if ``self`` is of finite type.

        This means that the cluster algebra associated to ``self`` has
        only a finite number of cluster variables.

        EXAMPLES::

            sage: mt = QuiverMutationType(['A', 2])
            sage: mt.is_finite()
            True

            sage: mt = QuiverMutationType(['A',[4, 2], 1])
            sage: mt.is_finite()
            False
        """
        return self._info['finite']

    def is_affine(self):
        """
        Return ``True`` if ``self`` is of affine type.

        EXAMPLES::

            sage: mt = QuiverMutationType(['A', 2])
            sage: mt.is_affine()
            False

            sage: mt = QuiverMutationType(['A',[4, 2], 1])
            sage: mt.is_affine()
            True
        """
        if self.is_irreducible():
            return self._info['affine']
        else:
            return False

    def is_elliptic(self):
        """
        Return ``True`` if ``self`` is of elliptic type.

        EXAMPLES::

            sage: mt = QuiverMutationType(['A', 2])
            sage: mt.is_elliptic()
            False

            sage: mt = QuiverMutationType(['E',6,[1, 1]])
            sage: mt.is_elliptic()
            True
        """
        if self.is_irreducible():
            return self._info['elliptic']
        else:
            return False

    def properties(self):
        """
        Print a scheme of all properties of ``self``.

        Most properties have natural definitions for either irreducible or
        reducible types.  ``affine`` and ``elliptic`` are only defined for
        irreducible types.

        EXAMPLES::

            sage: mut_type = QuiverMutationType(['A',3]); mut_type
            ['A', 3]
            sage: mut_type.properties()
            ['A', 3] has rank 3 and the following properties:
                - irreducible:       True
                - mutation finite:   True
                - simply-laced:      True
                - skew-symmetric:    True
                - finite:            True
                - affine:            False
                - elliptic:          False

            sage: mut_type = QuiverMutationType(['B',3]); mut_type
            ['B', 3]
            sage: mut_type.properties()
            ['B', 3] has rank 3 and the following properties:
                - irreducible:       True
                - mutation finite:   True
                - simply-laced:      False
                - skew-symmetric:    False
                - finite:            True
                - affine:            False
                - elliptic:          False

            sage: mut_type = QuiverMutationType(['B',3, 1]); mut_type
            ['BD', 3, 1]
            sage: mut_type.properties()
            ['BD', 3, 1] has rank 4 and the following properties:
                - irreducible:       True
                - mutation finite:   True
                - simply-laced:      False
                - skew-symmetric:    False
                - finite:            False
                - affine:            True
                - elliptic:          False

            sage: mut_type = QuiverMutationType(['E',6,[1, 1]]); mut_type
            ['E', 6, [1, 1]]
            sage: mut_type.properties()
            ['E', 6, [1, 1]] has rank 8 and the following properties:
                - irreducible:       True
                - mutation finite:   True
                - simply-laced:      False
                - skew-symmetric:    True
                - finite:            False
                - affine:            False
                - elliptic:          True

            sage: mut_type = QuiverMutationType(['A',3],['B',3]); mut_type
            [ ['A', 3], ['B', 3] ]
            sage: mut_type.properties()
            [ ['A', 3], ['B', 3] ] has rank 6 and the following properties:
                - irreducible:       False
                - mutation finite:   True
                - simply-laced:      False
                - skew-symmetric:    False
                - finite:            True

            sage: mut_type = QuiverMutationType('GR',[4,9]); mut_type
            ['GR', [4, 9]]
            sage: mut_type.properties()
            ['GR', [4, 9]] has rank 12 and the following properties:
                - irreducible:       True
                - mutation finite:   False
                - simply-laced:      True
                - skew-symmetric:    True
                - finite:            False
                - affine:            False
                - elliptic:          False
        """
        txt = '{} has rank {} and the following properties:'
        print(txt.format(self, self.rank()))
        s = "\t- {} {}"
        print(s.format('irreducible:      ', self.is_irreducible()))
        print(s.format('mutation finite:  ', self.is_mutation_finite()))
        print(s.format('simply-laced:     ', self.is_simply_laced()))
        print(s.format('skew-symmetric:   ', self.is_skew_symmetric()))
        print(s.format('finite:           ', self.is_finite()))
        if self.is_irreducible():
            print(s.format('affine:           ', self.is_affine()))
            print(s.format('elliptic:         ', self.is_elliptic()))


class QuiverMutationType_Irreducible(QuiverMutationType_abstract):
    """
    The mutation type for a cluster algebra or a quiver. Should not be
    called directly, but through :class:`QuiverMutationType`.
    """

    def __init__(self, letter, rank, twist=None):
        """
        Should not be called directly but through QuiverMutationType.

        INPUT:

        - ``letter`` -- the letter of the mutation type
        - ``rank`` -- the rank of the mutation type
        - ``twist`` -- the twist of the mutation type

        EXAMPLES::

            sage: QuiverMutationType('A',5)
            ['A', 5]

            sage: QuiverMutationType('A',[4,5], 1)
            ['A', [4, 5], 1]

            sage: QuiverMutationType('BB',5, 1)
            ['BB', 5, 1]

            sage: QuiverMutationType('X',6)
            ['X', 6]
        """
        # _rank and _bi_rank are initialized
        self._rank = None
        self._bi_rank = None

        # _graph and _digraph are initialized
        self._graph = Graph()
        self._digraph = DiGraph()

        # _info is initialized
        self._info = {}
        self._info['irreducible'] = True
        self._info['mutation_finite'] = False
        self._info['simply_laced'] = False
        self._info['skew_symmetric'] = False
        self._info['finite'] = False
        self._info['affine'] = False
        self._info['elliptic'] = False
        self._info['irreducible_components'] = False

        if isinstance(rank, tuple):
            rank = list(rank)
        if isinstance(twist, tuple):
            twist = list(twist)

        # _letter/twist is the input letter/twist
        self._letter = letter
        self._twist = twist

        data = [letter, rank, twist]

        # type A (finite and affine)
        if letter == 'A':
            if twist is None and rank in ZZ and rank > 0:
                self._rank = rank
                self._info['mutation_finite'] = True
                self._info['simply_laced'] = True
                self._info['skew_symmetric'] = True
                self._info['finite'] = True
            elif twist == 1 and isinstance(rank, list) and len(rank) == 2 and all(ri in ZZ and ri >= 0 for ri in rank) and rank != [0, 0]:
                if isinstance(rank, tuple):
                    rank = list(rank)
                    data[1] = rank
                rank = sorted(rank)
                self._bi_rank = rank
                self._rank = sum(self._bi_rank)
                self._info['mutation_finite'] = True
                if self._rank > 2:
                    self._info['simply_laced'] = True
                self._info['skew_symmetric'] = True
                if rank[0] > 0:
                    self._info['affine'] = True
                elif rank[0] == 0:
                    self._info['finite'] = True
            else:
                _mutation_type_error(data)
            # types ['A', 1] and ['A',[0, 1], 1] need to be treated on
            # itself (as there is no edge)
            if twist is None and self._rank == 1 or twist == 1 and self._rank == 1:
                self._graph.add_vertex(0)
            # type ['A',[1, 1], 1] needs to be treated on itself as well
            # (as there is a double edge)
            elif twist == 1 and self._bi_rank[0] == 1 and self._bi_rank[1] == 1:
                self._graph.add_edge(0, 1, 2)
            else:
                for i in range(self._rank - 1):
                    self._graph.add_edge(i, i + 1, 1)
                if twist == 1:
                    self._digraph.add_edge(self._rank - 1, 0, 1)
                    for i in range(self._rank - 1):
                        if i < (2 * self._bi_rank[0]) and i % 2 == 0:
                            self._digraph.add_edge(i + 1, i, 1)
                        else:
                            self._digraph.add_edge(i, i + 1, 1)

        # type B (finite)
        elif letter == 'B':
            if twist is None and rank in ZZ and rank > 1:
                self._rank = rank
                self._info['mutation_finite'] = True
                self._info['finite'] = True
            else:
                _mutation_type_error(data)
            for i in range(rank - 2):
                self._graph.add_edge(i, i+1, 1)
            if (rank % 2 == 0):
                self._graph.add_edge(rank-2, rank-1, (1, -2))
            else:
                self._graph.add_edge(rank-2, rank-1, (2, -1))

        # type C (finite)
        elif letter == 'C':
            if twist is None and rank in ZZ and rank > 1:
                self._rank = rank
                self._info['mutation_finite'] = True
                self._info['finite'] = True
            else:
                _mutation_type_error(data)
            for i in range(rank - 2):
                self._graph.add_edge(i, i+1, 1)
            if (rank % 2 == 0):
                self._graph.add_edge(rank-2, rank-1, (2, -1))
            else:
                self._graph.add_edge(rank-2, rank-1, (1, -2))

        # type BB (affine)
        elif letter == 'BB':
            if twist == 1 and rank in ZZ and rank > 1:
                self._rank = rank + 1
                self._info['mutation_finite'] = True
                self._info['affine'] = True
            else:
                _mutation_type_error(data)
            for i in range(rank - 2):
                self._graph.add_edge(i, i+1, 1)
            if rank % 2 == 0:
                self._graph.add_edge(rank-2, rank-1, (1, -2))
            else:
                self._graph.add_edge(rank-2, rank-1, (2, -1))
            self._graph.add_edge(rank, 0, (1, -2))

        # type CC (affine)
        elif letter == 'CC':
            if twist == 1 and rank in ZZ and rank > 1:
                self._rank = rank + 1
                self._info['mutation_finite'] = True
                self._info['affine'] = True
            else:
                _mutation_type_error(data)
            for i in range(rank - 2):
                self._graph.add_edge(i, i+1, 1)
            if rank % 2 == 0:
                self._graph.add_edge(rank-2, rank-1, (2, -1))
            else:
                self._graph.add_edge(rank-2, rank-1, (1, -2))
            self._graph.add_edge(rank, 0, (2, -1))

        # type BC (affine)
        elif letter == 'BC':
            if twist == 1 and rank in ZZ and rank >= 1:
                self._rank = rank + 1
                self._info['mutation_finite'] = True
                self._info['affine'] = True
            else:
                _mutation_type_error(data)
            if rank == 1:
                self._graph.add_edge(0, 1, (1, -4))
            else:
                for i in range(rank - 2):
                    self._graph.add_edge(i, i+1, 1)
                if (rank % 2 == 0):
                    self._graph.add_edge(rank-2, rank-1, (2, -1))
                else:
                    self._graph.add_edge(rank-2, rank-1, (1, -2))
                if twist == 1:
                    self._graph.add_edge(rank, 0, (1, -2))

        # type BD (affine)
        elif letter == 'BD':
            if twist == 1 and rank in ZZ and rank > 2:
                self._rank = rank + 1
                self._info['mutation_finite'] = True
                self._info['affine'] = True
            else:
                _mutation_type_error(data)
            for i in range(rank - 2):
                self._graph.add_edge(i, i+1, 1)
            if (rank % 2 == 0):
                self._graph.add_edge(rank-2, rank-1, (1, -2))
            else:
                self._graph.add_edge(rank-2, rank-1, (2, -1))
            if twist == 1:
                self._graph.add_edge(rank, 1, 1)

        # type CD (affine)
        elif letter == 'CD':
            if twist == 1 and rank in ZZ and rank > 2:
                self._rank = rank + 1
                self._info['mutation_finite'] = True
                self._info['affine'] = True
            else:
                _mutation_type_error(data)
            for i in range(rank - 2):
                self._graph.add_edge(i, i+1, 1)
            if (rank % 2 == 0):
                self._graph.add_edge(rank-2, rank-1, (2, -1))
            else:
                self._graph.add_edge(rank-2, rank-1, (1, -2))
            if twist == 1:
                self._graph.add_edge(rank, 1, 1)

        # type D (finite and affine)
        elif letter == 'D':
            if rank in ZZ and rank > 3 and twist is None:
                self._rank = rank
                self._info['mutation_finite'] = True
                self._info['simply_laced'] = True
                self._info['skew_symmetric'] = True
                self._info['finite'] = True
            elif twist == 1 and rank in ZZ and rank > 3:
                self._rank = rank + 1
                self._info['mutation_finite'] = True
                self._info['simply_laced'] = True
                self._info['skew_symmetric'] = True
                self._info['affine'] = True
            else:
                _mutation_type_error(data)
            for i in range(rank - 2):
                self._graph.add_edge(i, i+1, 1)

            self._graph.add_edge(rank-3, rank-1, 1)
            if twist is not None:
                self._graph.add_edge(rank, 1, 1)

        # type E (finite, affine and elliptic)
        elif letter == 'E':
            if rank in [6, 7, 8] and twist is None:
                self._rank = rank
                self._info['mutation_finite'] = True
                self._info['simply_laced'] = True
                self._info['skew_symmetric'] = True
                self._info['finite'] = True
                if rank == 6:
                    self._graph.add_edges([(0, 1), (1, 2), (2, 3), (3, 4), (2, 5)])
                elif rank == 7:
                    self._graph.add_edges([(0, 1), (1, 2), (2, 3),
                                           (3, 4), (4, 5), (2, 6)])
                elif rank == 8:
                    self._graph.add_edges([(0, 1), (1, 2), (2, 3),
                                           (3, 4), (4, 5), (5, 6), (2, 7)])
            elif rank in [6, 7, 8] and twist == 1:
                self._rank = rank + 1
                self._info['mutation_finite'] = True
                self._info['simply_laced'] = True
                self._info['skew_symmetric'] = True
                self._info['affine'] = True
                if rank == 6:
                    self._graph.add_edges([(0, 1), (1, 2), (2, 3), (3, 4), (2, 5), (5, 6)])
                elif rank == 7:
                    self._graph.add_edges([(0, 1), (1, 2), (2, 3), (3, 4), (4, 5), (5, 6), (3, 7)])
                elif rank == 8:
                    self._graph.add_edges([(0, 1), (1, 2), (2, 3), (3, 4), (4, 5), (5, 6), (6, 7), (2, 8)])
            elif rank in [6, 7, 8] and twist == [1, 1]:
                self._rank = rank + 2
                self._info['mutation_finite'] = True
                self._info['skew_symmetric'] = True
                self._info['elliptic'] = True
                if rank == 6:
                    self._digraph.add_edges([(0, 1, 1), (1, 2, 1), (3, 2, 1), (3, 4, 1), (5, 6, 1), (6, 7, 1), (5, 1, 1), (2, 5, 2), (5, 3, 1), (6, 2, 1)])
                elif rank == 7:
                    self._digraph.add_edges([(1, 0, 1), (1, 2, 1), (2, 3, 1), (4, 3, 1), (4, 5, 1),
                                             (6, 5, 1), (7, 8, 1), (3, 7, 2), (7, 2, 1), (7, 4, 1), (8, 3, 1)])
                elif rank == 8:
                    self._digraph.add_edges([(0, 1, 1), (1, 9, 1), (3, 9, 1), (3, 4, 1), (2, 8, 1), (2, 1, 1),
                                             (9, 2, 2), (2, 3, 1), (8, 9, 1), (5, 4, 1), (5, 6, 1), (7, 6, 1)])
            # type E (mutation infinite)
            elif rank > 9 and twist is None:
                self._info['simply_laced'] = True
                self._info['skew_symmetric'] = True
                self._rank = rank
                for i in range(rank-2):
                    self._graph.add_edge(i, i+1, 1)
                self._graph.add_edge(2, rank-1)
            else:
                _mutation_type_error(data)

        # type AE (mutation infinite)
        elif letter == 'AE':
            if isinstance(rank, list) and len(rank) == 2 and all(rank[i] in ZZ and rank[i] > 0 for i in [0, 1]) and twist is None:
                if isinstance(rank, tuple):
                    rank = list(rank)
                    data[1] = rank
                rank = sorted(rank)
                self._bi_rank = rank
                self._rank = sum(self._bi_rank) + 1
                if self._rank > 3:
                    self._info['simply_laced'] = True
                self._info['skew_symmetric'] = True
                if self._bi_rank == [1, 1]:
                    self._graph.add_edges([(0, 1, 2), (1, 2, None)])
                else:
                    self._digraph.add_edge(self._rank - 2, 0)
                    for i in range(self._rank-2):
                        if i < (2 * self._bi_rank[0]) and i % 2 == 0:
                            self._digraph.add_edge(i + 1, i)
                        else:
                            self._digraph.add_edge(i, i + 1)
                    self._digraph.add_edge(self._rank-2, self._rank-1)
            else:
                _mutation_type_error(data)

        # type BE (mutation infinite)
        elif letter == 'BE':
            if rank > 4 and twist is None:
                self._rank = rank
                for i in range(rank-3):
                    self._graph.add_edge(i, i+1)
                self._graph.add_edge(2, rank-1)
                if rank % 2 == 0:
                    self._graph.add_edge(rank-3, rank-2, (2, -1))
                else:
                    self._graph.add_edge(rank-3, rank-2, (1, -2))
            else:
                _mutation_type_error(data)

        # type CE (mutation infinite)
        elif letter == 'CE':
            if rank > 4 and twist is None:
                self._rank = rank
                for i in range(rank-3):
                    self._graph.add_edge(i, i+1)
                self._graph.add_edge(2, rank-1)
                if rank % 2 == 0:
                    self._graph.add_edge(rank-3, rank-2, (1, -2))
                else:
                    self._graph.add_edge(rank-3, rank-2, (2, -1))
            else:
                _mutation_type_error(data)

        # type DE (mutation infinite)
        elif letter == 'DE':
            if rank > 5 and twist is None:
                self._rank = rank
                self._info['simply_laced'] = True
                self._info['skew_symmetric'] = True
                for i in range(rank-3):
                    self._graph.add_edge(i, i+1)
                self._graph.add_edge(2, rank-2)
                self._graph.add_edge(rank-4, rank-1)
            else:
                _mutation_type_error(data)

        # type F (finite, affine, and elliptic)
        elif letter == 'F':
            if rank == 4 and twist is None:
                self._rank = rank
                self._info['mutation_finite'] = True
                self._info['finite'] = True
                self._graph.add_edges([(0, 1, None), (1, 2, (2, -1)), (2, 3, None)])
            elif rank == 4 and twist == 1:
                self._rank = rank + 1
                self._info['mutation_finite'] = True
                self._info['affine'] = True
                self._graph.add_edges([(0, 1, None), (1, 2, None),
                                       (2, 3, (1, -2)), (3, 4, None)])
            elif rank == 4 and twist == -1:
                self._rank = rank + 1
                self._info['mutation_finite'] = True
                self._info['affine'] = True
                self._graph.add_edges([(0, 1, None), (1, 2, None),
                                       (2, 3, (2, -1)), (3, 4, None)])
            elif rank == 4 and (twist == [1, 2]):
                self._rank = rank + 2
                self._info['mutation_finite'] = True
                self._info['elliptic'] = True
                self._digraph.add_edges([(0, 1, None), (1, 2, None),
                                         (2, 3, (2, -1)), (4, 2, (1, -2)),
                                         (3, 4, 2), (4, 5, None), (5, 3, None)])
            elif rank == 4 and (twist == [2, 1]):
                self._rank = rank + 2
                self._info['mutation_finite'] = True
                self._info['elliptic'] = True
                self._digraph.add_edges([(0, 1, None), (1, 2, None),
                                         (2, 3, (1, -2)), (4, 2, (2, -1)),
                                         (3, 4, 2), (4, 5, None), (5, 3, None)])
            elif rank == 4 and twist == [2, 2]:
                self._rank = rank + 2
                self._info['mutation_finite'] = True
                self._info['elliptic'] = True
                self._digraph.add_edges([(0, 1, None), (1, 2, None),
                                         (3, 1, None), (2, 3, 2),
                                         (4, 2, (2, -1)), (3, 4, (1, -2)),
                                         (5, 4, None)])
            elif rank == 4 and twist == [1, 1]:
                self._rank = rank + 2
                self._info['mutation_finite'] = True
                self._info['elliptic'] = True
                self._digraph.add_edges([(0, 1, None), (1, 2, None),
                                         (3, 1, None), (2, 3, 2), (4, 2, (1, -2)),
                                         (3, 4, (2, -1)), (5, 4, None)])
            else:
                _mutation_type_error(data)

        # type G (finite, affine, and elliptic)
        elif letter == 'G':
            if rank == 2 and twist is None:
                self._rank = rank
                self._info['mutation_finite'] = True
                self._info['finite'] = True
                self._graph.add_edges([(0, 1, (1, -3))])
            elif rank == 2 and twist == -1:
                self._rank = rank + 1
                self._info['mutation_finite'] = True
                self._info['affine'] = True
                self._graph.add_edges([(0, 1, None), (1, 2, (1, -3))])
            elif rank == 2 and twist == 1:
                self._rank = rank + 1
                self._info['mutation_finite'] = True
                self._info['affine'] = True
                self._graph.add_edges([(0, 1, None), (1, 2, (3, -1))])
            elif rank == 2 and (twist == [1, 3]):
                self._rank = rank + 2
                self._info['mutation_finite'] = True
                self._info['elliptic'] = True
                self._digraph.add_edges([(0, 1, None), (1, 2, (3, -1)),
                                         (3, 1, (1, -3)), (2, 3, 2)])
            elif rank == 2 and (twist == [3, 1]):
                self._rank = rank + 2
                self._info['mutation_finite'] = True
                self._info['elliptic'] = True
                self._digraph.add_edges([(0, 1, None), (1, 2, (1, -3)),
                                         (3, 1, (3, -1)), (2, 3, 2)])
            elif rank == 2 and twist == [3, 3]:
                self._rank = rank + 2
                self._info['mutation_finite'] = True
                self._info['elliptic'] = True
                self._digraph.add_edges([(1, 0, None), (0, 2, 2), (3, 0, (3, -1)),
                                         (2, 1, None), (2, 3, (1, -3))])
            elif rank == 2 and twist == [1, 1]:
                self._rank = rank + 2
                self._info['mutation_finite'] = True
                self._info['elliptic'] = True
                self._digraph.add_edges([(1, 0, None), (0, 2, 2), (3, 0, (1, -3)),
                                         (2, 1, None), (2, 3, (3, -1))])
            else:
                _mutation_type_error(data)

        # type GR (mutation infinite)
        elif letter == 'GR':
            if twist is None and isinstance(rank, list) and len(rank) == 2 and all(rank[i] in ZZ and rank[i] > 0 for i in [0, 1]) and rank[1] - 1 > rank[0] > 1:
                gr_rank = (rank[0]-1, rank[1]-rank[0]-1)
                self._rank = prod(gr_rank)
                self._info['simply_laced'] = True
                self._info['skew_symmetric'] = True
                a, b = gr_rank
                for i in range(a):
                    for j in range(b):
                        if i < a-1:
                            if (i+j) % 2 == 0:
                                self._digraph.add_edge(i*b+j, (i+1)*b+j)
                            else:
                                self._digraph.add_edge((i+1)*b+j, i*b+j)
                        if j < b-1:
                            if (i+j) % 2 == 0:
                                self._digraph.add_edge(i*b+j+1, i*b+j)
                            else:
                                self._digraph.add_edge(i*b+j, i*b+j+1)
            else:
                _mutation_type_error(data)

        # type R2 (rank 2 finite mutation types)
        elif letter == 'R2':
            if twist is None and isinstance(rank, list) and len(rank) == 2 and all(rank[i] in ZZ and rank[i] > 0 for i in [0, 1]):
                rank = sorted(rank)
                b, c = rank
                self._rank = 2
                if b == c:
                    self._info['skew_symmetric'] = True
                self._graph.add_edge(0, 1, (b, -c))
            else:
                _mutation_type_error(data)

        # type T
        elif letter == 'T':
            if twist is None and isinstance(rank, list) and len(rank) == 3 and all(rank[i] in ZZ and rank[i] > 0 for i in [0, 1, 2]):
                if isinstance(rank, tuple):
                    rank = list(rank)
                    data[1] = rank
                rank = sorted(rank)
                self._rank = sum(rank) - 2
                self._info['simply_laced'] = True
                self._info['skew_symmetric'] = True
                r, p, q = rank
                for i in range(q-1):
                    if i == 0:
                        self._graph.add_edge(0, 1)
                        self._graph.add_edge(0, r)
                        self._graph.add_edge(0, r+p-1)
                    else:
                        if i < r-1:
                            self._graph.add_edge(i, i+1)
                        if i < p-1:
                            self._graph.add_edge(i+r-1, i+r)
                        self._graph.add_edge(i+r+p-2, i+r+p-1)
            else:
                _mutation_type_error(data)

        # type TR (mutation infinite if rank > 2)
        elif letter == 'TR':
            # type ['TR', 1] needs to be treated on itself (as there is no edge)
            if twist is None and rank == 1:
                self._graph.add_vertex(0)
            elif twist is None and rank > 1:
                self._rank = rank*(rank+1)//2
                self._info['simply_laced'] = True
                self._info['skew_symmetric'] = True
                level = 0
                while level < rank:
                    nr = rank*level-sum(range(level))
                    for i in range(nr, nr+rank-level-1):
                        self._digraph.add_edge(i, i+1)
                        self._digraph.add_edge(i+rank-level, i)
                        self._digraph.add_edge(i+1, i+rank-level)
                    level += 1
            else:
                _mutation_type_error(data)

        # type X
        elif letter == 'X':
            if rank in [6, 7] and twist is None:
                self._rank = rank
                self._info['mutation_finite'] = True
                self._info['skew_symmetric'] = True
                self._digraph.add_edges([(0, 1, 2), (1, 2, None), (2, 0, None),
                                         (2, 3, None), (3, 4, 2), (4, 2, None),
                                         (2, 5, None)])
                if rank == 7:
                    self._digraph.add_edges([(5, 6, 2), (6, 2, None)])
            else:
                _mutation_type_error(data)

        # otherwise, an error is raised
        else:
            _mutation_type_error(data)

        # in the bipartite case, the digraph is constructed from the graph
        if not self._digraph:
            if self._graph.is_bipartite():
                self._digraph = _bipartite_graph_to_digraph(self._graph)
            else:
                raise ValueError('The QuiverMutationType does not have '
                                 'a Coxeter diagram.')

        # in the other cases, the graph is constructed from the digraph
        if not self._graph:
            self._graph = self._digraph.to_undirected()

        # _description is as for CartanType
        if twist:
            self._description = str([letter, rank, twist])
        else:
            self._description = str([letter, rank])

    def irreducible_components(self):
        """
        Return a list of all irreducible components of ``self``.

        EXAMPLES::

            sage: mut_type = QuiverMutationType('A',3); mut_type
            ['A', 3]
            sage: mut_type.irreducible_components()
            (['A', 3],)
        """
        return (self,)

    @cached_method
    def class_size(self):
        r"""
        If it is known, the size of the mutation class of all quivers
        which are mutation equivalent to the standard quiver of
        ``self`` (up to isomorphism) is returned.

        Otherwise, :obj:`NotImplemented` is returned.

        Formula for finite type A is taken from Torkildsen - Counting
        cluster-tilted algebras of type `A_n`.
        Formulas for affine type A and finite type D are taken from Bastian,
        Prellberg, Rubey, Stump - Counting the number of elements in the
        mutation classes of `\widetilde A_n` quivers.
        Formulas for finite and affine types B and C are
        proven but not yet published.
        Conjectural formulas for several other non-simply-laced affine types
        are implemented.
        Exceptional Types (finite, affine, and elliptic) E, F, G, and X are
        hardcoded.

        EXAMPLES::

            sage: mut_type = QuiverMutationType( ['A',5] ); mut_type
            ['A', 5]
            sage: mut_type.class_size()
            19

            sage: mut_type = QuiverMutationType( ['A',[10,3], 1] ); mut_type
            ['A', [3, 10], 1]
            sage: mut_type.class_size()
            142120

            sage: mut_type = QuiverMutationType( ['B',6] ); mut_type
            ['B', 6]
            sage: mut_type.class_size()
            132

            sage: mut_type = QuiverMutationType( ['BD',6, 1] ); mut_type
            ['BD', 6, 1]
            sage: mut_type.class_size()
            Warning: This method uses a formula which has not been proved correct.
            504

        Check that :issue:`14048` is fixed::

            sage: mut_type = QuiverMutationType( ['F',4,(2, 1)] )
            sage: mut_type.class_size()
            90
        """
        if not self.is_mutation_finite():
            return infinity

        # type A (finite and affine)
        if self._letter == 'A':
            # the formula is taken from Torkildsen - Counting
            # cluster-tilted algebras of type A
            if self.is_finite():
                n = self._rank
                a = binomial(2*(n+1), n+1) // (n+2)
                if n % 2 == 1:
                    a += binomial(n+1, (n+1)//2)
                if n % 3 == 0:
                    a += 2 * binomial(2*n//3, n//3)
                return a // (n+3)
            # the formula is taken from Bastian, Prellberg, Rubey, Stump
            elif self.is_affine():
                i, j = self._bi_rank
                i = ZZ(i)
                j = ZZ(j)
                n = i+j
                f = euler_phi
                if i == j:
                    return (binomial(2 * i, i) +
                            sum(f(k) * binomial(2 * i // k, i // k)**2
                                for k in i.divisors()
                                if k in j.divisors()) // n) // 4
                else:
                    return sum(f(k) * binomial(2 * i // k, i // k) *
                               binomial(2 * j // k, j // k)
                               for k in i.divisors()
                               if k in j.divisors()) // (2 * n)

        # types B and C (finite and affine)
        elif self._letter in ['B', 'C']:
            # this formula is proven but nowhere published correctness
            # is clear enough that I don't think a warning is needed
            if self.is_finite():
                n = self._rank
                return binomial(2 * n, n) // (n + 1)

        elif self._letter in ['BB', 'CC']:
            # these two formulas are not yet proven
            print("Warning: This method uses a formula "
                  "which has not been proved correct.")
            if self.is_affine():
                if self._twist == 1:
                    n = self._rank - 1
                    N = binomial(2 * n - 1, n - 1)
                    if n % 2:
                        return N
                    else:
                        return N + binomial(n - 1, n // 2 - 1)

        # type BC (affine)
        elif self._letter == 'BC':
            # this formula is not yet proven
            print("Warning: This method uses a formula "
                  "which has not been proved correct.")
            if self.is_affine():
                if self._twist == 1:
                    n = self._rank - 1
                    return binomial(2 * n, n)

        # types BD and CD (affine)
        elif self._letter in ['BD', 'CD']:
            # this formula is not yet proven
            print("Warning: This method uses a formula "
                  "which has not been proved correct.")
            if self.is_affine():
                if self._twist == 1:
                    n = self._rank - 2
                    return 2 * binomial(2 * n, n)

        # type D (finite and affine)
        elif self._letter == 'D':
            # the formula is taken from Bastian, Prellberg, Rubey, Stump
            if self.is_finite():
                if self._rank == 4:
                    return 6
                else:
                    f = euler_phi
                    n = ZZ(self._rank)
                    return sum(f(n // k) * binomial(2 * k, k)
                               for k in n.divisors()) // (2 * n)
            # this formula is not yet proven
            elif self.is_affine():
                n = self._rank - 3
                if n == 2:
                    return 9
                else:
                    print("Warning: This method uses a formula "
                          "which has not been proved correct.")
                    if n % 2:
                        return 2 * binomial(2 * n, n)
                    else:
                        return 2 * binomial(2 * n, n) + binomial(n, n // 2)

        # the exceptional types are hard-coded
        # type E (finite, affine and elliptic)
        elif self._letter == 'E':
            if self.is_finite():
                if self._rank == 6:
                    return 67
                elif self._rank == 7:
                    return 416
                elif self._rank == 8:
                    return 1574
            elif self.is_affine():
                if self._rank == 7:
                    return 132
                elif self._rank == 8:
                    return 1080
                elif self._rank == 9:
                    return 7560
            elif self.is_elliptic():
                if self._rank == 8:
                    return 49
                elif self._rank == 9:
                    return 506
                elif self._rank == 10:
                    return 5739

        # type F
        elif self._letter == 'F':
            if self.is_finite():
                return 15
            elif self.is_affine():
                return 60
            elif self.is_elliptic():
                if self._twist == [1, 2]:
                    return 90
                if self._twist == [1, 1] or self._twist == [2, 2]:
                    return 35

        # type G
        elif self._letter == 'G':
            if self.is_finite():
                return 2
            elif self.is_affine():
                return 6
            elif self.is_elliptic():
                if self._twist == [1, 3]:
                    return 7
                if self._twist == [1, 1] or self._twist == [3, 3]:
                    return 2

        # type X
        elif self._letter == 'X':
            if self._rank == 6:
                return 5
            elif self._rank == 7:
                return 2

        # otherwise the size is returned to be unknown
        else:
            print("Size unknown")
            return NotImplemented

    def dual(self):
        """
        Return the :class:`QuiverMutationType` which is dual to ``self``.

        EXAMPLES::

            sage: mut_type = QuiverMutationType('A',5); mut_type
            ['A', 5]
            sage: mut_type.dual()
            ['A', 5]

            sage: mut_type = QuiverMutationType('B',5); mut_type
            ['B', 5]
            sage: mut_type.dual()
            ['C', 5]
            sage: mut_type.dual().dual()
            ['B', 5]
            sage: mut_type.dual().dual() == mut_type
            True
        """
        letter = self.letter()
        # the self-dual cases
        if letter != 'BC' and letter[0] in ['B', 'C']:
            if letter == 'BB':
                letter = 'CC'
            elif letter == 'CC':
                letter = 'BB'
            elif letter[0] == 'B':
                letter = 'C' + letter[1:]
            elif letter[0] == 'C':
                letter = 'B' + letter[1:]
            rank = self._rank
            if self.is_affine():
                rank -= 1
            twist = self._twist
            return QuiverMutationType(letter, rank, twist)
        # the cases F and G have non-trivial duality in some cases
        elif letter in ['F', 'G']:
            if self.is_finite():
                return self
            elif self.is_affine():
                rank = self._rank - 1
                twist = - self._twist
            elif self.is_elliptic():
                twist = self._twist
                rank = self._rank - 2
                if letter == 'F':
                    if self._twist == [2, 2]:
                        twist == [1, 1]
                    if self._twist == [1, 1]:
                        twist == [2, 2]
                if letter == 'G':
                    if self._twist == [3, 3]:
                        twist = [1, 1]
                    elif self._twist == [1, 1]:
                        twist = [3, 3]
            else:
                rank = self._rank
            return QuiverMutationType(letter, rank, twist)
        else:
            return self


class QuiverMutationType_Reducible(QuiverMutationType_abstract):
    """
    The mutation type for a cluster algebra or a quiver. Should not be
    called directly, but through :class:`QuiverMutationType`.  Inherits from
    :class:`QuiverMutationType_abstract`.
    """

    def __init__(self, *args):
        """
        Should not be called directly, but through QuiverMutationType.

        INPUT:

        - ``data`` -- list; each of whose entries is a
          :class:`QuiverMutationType_Irreducible`

        EXAMPLES::

            sage: QuiverMutationType(['A',4],['B',6])
            [ ['A', 4], ['B', 6] ]
        """
        data = args
        if len(data) < 2 or not all(isinstance(comp, QuiverMutationType_Irreducible) for comp in data):
            _mutation_type_error(data)

        # _info is initialized
        self._info = {}
        self._info['irreducible'] = False
        self._info['mutation_finite'] = all(comp.is_mutation_finite()
                                            for comp in data)
        self._info['simply_laced'] = all(comp.is_simply_laced()
                                         for comp in data)
        self._info['skew_symmetric'] = all(comp.is_skew_symmetric()
                                           for comp in data)
        self._info['finite'] = all(comp.is_finite() for comp in data)
        self._info['irreducible_components'] = copy(data)

        #  letter and rank are initialized
        self._letter = ''
        self._rank = 0

        # graph and digraph are initialized
        self._graph = Graph()
        self._digraph = DiGraph()

        for comp in data:
            if self._letter:
                self._letter += ' x '
            self._letter += comp._letter
            self._rank += comp._rank
            self._graph = self._graph.disjoint_union(comp._graph,
                                                     labels='integers')
            self._digraph = self._digraph.disjoint_union(comp._digraph,
                                                         labels='integers')
        self._graph.name('')
        self._digraph.name('')

        # _description is as for CartanType
        self._description = "[ "
        comps = self.irreducible_components()
        for i in range(len(comps)):
            if i > 0:
                self._description += ", "
            self._description += comps[i]._description
        self._description += " ]"

    def irreducible_components(self):
        """
        Return a list of all irreducible components of ``self``.

        EXAMPLES::

            sage: mut_type = QuiverMutationType('A',3); mut_type
            ['A', 3]
            sage: mut_type.irreducible_components()
            (['A', 3],)

            sage: mut_type = QuiverMutationType(['A',3],['B',3]); mut_type
            [ ['A', 3], ['B', 3] ]
            sage: mut_type.irreducible_components()
            (['A', 3], ['B', 3])

            sage: mut_type = QuiverMutationType(['A',3],['B',3],['X',6])
            sage: mut_type
            [ ['A', 3], ['B', 3], ['X', 6] ]
            sage: mut_type.irreducible_components()
            (['A', 3], ['B', 3], ['X', 6])
        """
        return self._info['irreducible_components']

    @cached_method
    def class_size(self):
        """
        If it is known, the size of the mutation class of all quivers
        which are mutation equivalent to the standard quiver of
        ``self`` (up to isomorphism) is returned.

        Otherwise, :obj:`NotImplemented` is returned.

        EXAMPLES::

            sage: mut_type = QuiverMutationType(['A',3],['B',3]); mut_type
            [ ['A', 3], ['B', 3] ]
            sage: mut_type.class_size()
            20

            sage: mut_type = QuiverMutationType(['A',3],['B',3],['X',6])
            sage: mut_type
            [ ['A', 3], ['B', 3], ['X', 6] ]
            sage: mut_type.class_size()
            100
        """
        if not self.is_mutation_finite():
            return infinity

        components = []
        multiplicities = []
        for x in self.irreducible_components():
            if not components.count(x):
                components.append(x)
                multiplicities.append(1)
            else:
                y = components.index(x)
                multiplicities[y] += 1

        sizes = [x.class_size() for x in components]
        if NotImplemented in sizes:
            print("Size unknown")
            return NotImplemented
        else:
            return prod(binomial(sizes[i] + multiplicities[i] - 1,
                                 multiplicities[i])
                        for i in range(len(sizes)))

    def dual(self):
        """
        Return the :class:`QuiverMutationType` which is dual to ``self``.

        EXAMPLES::

            sage: mut_type = QuiverMutationType(['A',5],['B',6],['C',5],['D',4]); mut_type
            [ ['A', 5], ['B', 6], ['C', 5], ['D', 4] ]
            sage: mut_type.dual()
            [ ['A', 5], ['C', 6], ['B', 5], ['D', 4] ]
        """
        comps = self.irreducible_components()
        return QuiverMutationType([comp.dual() for comp in comps])


def _construct_classical_mutation_classes(n) -> dict[tuple, list | set]:
    r"""
    Return a dict with keys being tuples representing regular
    QuiverMutationTypes of the given rank, and with values being lists
    or sets containing all mutation equivalent quivers as dig6 data.

    EXAMPLES::

        sage: from sage.combinat.cluster_algebra_quiver.quiver_mutation_type import _construct_classical_mutation_classes
        sage: rank_2_classes = _construct_classical_mutation_classes(2) # long time
        sage: for mut_class in sorted(rank_2_classes.keys(),key=str): # long time
        ....:   print("{} {}".format(mut_class, rank_2_classes[mut_class]))
        ('A', (1, 1), 1) [('AO', (((0, 1), (2, -2)),))]
        ('A', 2) [('AO', ())]
        ('B', 2) [('AO', (((0, 1), (1, -2)),)), ('AO', (((0, 1), (2, -1)),))]
        ('BC', 1, 1) [('AO', (((0, 1), (1, -4)),)),
        ('AO', (((0, 1), (4, -1)),))]
    """
    from sage.combinat.cluster_algebra_quiver.quiver import ClusterQuiver
    data: dict[tuple, set | list] = {}

    # finite A
    data[('A', n)] = ClusterQuiver(['A', n]).mutation_class(data_type='dig6')
    # affine A
    for j in range(1, n//2+1):
        data[('A', (n-j, j), 1)] = ClusterQuiver(['A', [n-j, j], 1]).mutation_class(data_type='dig6')
    # finite B
    if n > 1:
        data[('B', n)] = ClusterQuiver(['B', n]).mutation_class(data_type='dig6')
    # affine B
    if n > 2:
        data[('BB', n-1, 1)] = ClusterQuiver(['BB', n-1, 1]).mutation_class(data_type='dig6')
    # finite C
    if n > 2:
        data[('C', n)] = ClusterQuiver(['C', n]).mutation_class(data_type='dig6')
    # affine C
    if n > 1:
        data[('BC', n-1, 1)] = ClusterQuiver(['BC', n-1, 1]).mutation_class(data_type='dig6')
    # affine CC
    if n > 2:
        data[('CC', n-1, 1)] = ClusterQuiver(['CC', n-1, 1]).mutation_class(data_type='dig6')
    # affine BD
    if n > 3:
        data[('BD', n-1, 1)] = ClusterQuiver(['BD', n-1, 1]).mutation_class(data_type='dig6')
    # affine CD
    if n > 3:
        data[('CD', n-1, 1)] = ClusterQuiver(['CD', n-1, 1]).mutation_class(data_type='dig6')
    # finite D
    if n > 3:
        data[('D', n)] = ClusterQuiver(['D', n]).mutation_class(data_type='dig6')
    # affine D
    if n > 4:
        data[('D', n-1, 1)] = ClusterQuiver(['D', n-1, 1]).mutation_class(data_type='dig6')

    return data


def _construct_exceptional_mutation_classes(n) -> dict[tuple, list | set]:
    r"""
    Return a dict with keys being tuples representing exceptional
    QuiverMutationTypes of the given rank, and with values being lists
    or sets containing all mutation equivalent quivers as dig6 data.

    EXAMPLES::

        sage: from sage.combinat.cluster_algebra_quiver.quiver_mutation_type import _construct_exceptional_mutation_classes
        sage: rank_3_exceptional = _construct_exceptional_mutation_classes(3) # long time
        sage: for mut_class in sorted(rank_3_exceptional.keys(), key=str): # long time
        ....:   print("{} {}".format(mut_class, rank_3_exceptional[mut_class]))
        ('G', 2, -1) [('BH?', (((1, 2), (1, -3)),)),
        ('BGO', (((2, 1), (3, -1)),)), ('BW?', (((0, 1), (3, -1)),)),
        ('BP?', (((0, 1), (1, -3)),)),
        ('BP_', (((0, 1), (1, -3)), ((2, 0), (3, -1)))),
        ('BP_', (((0, 1), (3, -1)), ((1, 2), (1, -3)), ((2, 0), (2, -2))))]
        ('G', 2, 1) [('BH?', (((1, 2), (3, -1)),)),
        ('BGO', (((2, 1), (1, -3)),)), ('BW?', (((0, 1), (1, -3)),)),
        ('BP?', (((0, 1), (3, -1)),)),
        ('BKO', (((1, 0), (3, -1)), ((2, 1), (1, -3)))),
        ('BP_', (((0, 1), (2, -2)), ((1, 2), (1, -3)), ((2, 0), (3, -1))))]
    """
    from sage.combinat.cluster_algebra_quiver.quiver import ClusterQuiver
    data: dict[tuple, list | set] = {}
    # finite E
    if n in [6, 7, 8]:
        data[('E', n)] = ClusterQuiver(['E', n]).mutation_class(data_type='dig6')
    # affine E
    if n in [7, 8, 9]:
        data[('E', n - 1, 1)] = ClusterQuiver(['E', n - 1, 1]).mutation_class(data_type='dig6')
    # elliptic E
    if n in [8, 9, 10]:
        data[('E', n - 2, (1, 1))] = ClusterQuiver(['E', n - 2, [1, 1]]).mutation_class(data_type='dig6')
    # finite F
    if n == 4:
        data[('F', 4)] = ClusterQuiver(['F', 4]).mutation_class(data_type='dig6')
    # affine F
    if n == 5:
        data[('F', 4, 1)] = ClusterQuiver(['F', 4, 1]).mutation_class(data_type='dig6')
        data[('F', 4, -1)] = ClusterQuiver(['F', 4, -1]).mutation_class(data_type='dig6')
    # finite G
    if n == 2:
        data[('G', 2)] = ClusterQuiver(['G', 2]).mutation_class(data_type='dig6')
    # affine G
    if n == 3:
        data[('G', 2, 1)] = ClusterQuiver(['G', 2, 1]).mutation_class(data_type='dig6')
        data[('G', 2, -1)] = ClusterQuiver(['G', 2, -1]).mutation_class(data_type='dig6')
    # elliptic G
    if n == 4:
        data[('G', 2, (1, 3))] = ClusterQuiver(['G', 2, (1, 3)]).mutation_class(data_type='dig6')
        data[('G', 2, (1, 1))] = ClusterQuiver(['G', 2, (1, 1)]).mutation_class(data_type='dig6')
        data[('G', 2, (3, 3))] = ClusterQuiver(['G', 2, (3, 3)]).mutation_class(data_type='dig6')
    # X
    if n in [6, 7]:
        data[('X', n)] = ClusterQuiver(['X', n]).mutation_class(data_type='dig6')
    # elliptic F
    if n == 6:
        data[('F', 4, (1, 2))] = ClusterQuiver(['F', 4, (1, 2)]).mutation_class(data_type='dig6')
        data[('F', 4, (1, 1))] = ClusterQuiver(['F', 4, (1, 1)]).mutation_class(data_type='dig6')
        data[('F', 4, (2, 2))] = ClusterQuiver(['F', 4, (2, 2)]).mutation_class(data_type='dig6')

    return data


def _save_data_dig6(n, types='ClassicalExceptional', verbose=False):
    """
    Save all exceptional mutation classes as dig6 data into the file ``exc_classes_n.dig6`` in the folder ``DOT_SAGE``.

    TESTS::

        sage: # needs sage.modules
        sage: from sage.combinat.cluster_algebra_quiver.quiver_mutation_type import save_quiver_data
        sage: save_quiver_data(2)  # indirect doctest
        <BLANKLINE>
        The following types are saved to file ... and will now be used to determine quiver mutation types:
        [('A', 1)]
        <BLANKLINE>
        The following types are saved to file ... and will now be used to determine quiver mutation types:
        [('A', (1, 1), 1), ('A', 2), ('B', 2), ('BC', 1, 1), ('G', 2)]
        sage: save_quiver_data(2, up_to=False)  # indirect doctest
        <BLANKLINE>
        The following types are saved to file ... and will now be used to determine quiver mutation types:
        [('A', (1, 1), 1), ('A', 2), ('B', 2), ('BC', 1, 1), ('G', 2)]
        sage: save_quiver_data(2, up_to=False, types='Classical')  # indirect doctest
        <BLANKLINE>
        The following types are saved to file ... and will now be used to determine quiver mutation types:
        [('A', (1, 1), 1), ('A', 2), ('B', 2), ('BC', 1, 1)]
        sage: save_quiver_data(2, up_to=False, types='Exceptional')  # indirect doctest
        <BLANKLINE>
        The following types are saved to file ... and will now be used to determine quiver mutation types:
        [('G', 2)]
        sage: save_quiver_data(2, up_to=False, verbose=False)  # indirect doctest
    """
    data = {}
    possible_types = ['Classical', 'ClassicalExceptional', 'Exceptional']
    if types not in possible_types:
        raise ValueError('The third input must be either ClassicalExceptional'
                         ' (default), Classical, or Exceptional.')

    if types in possible_types[:2]:
        data.update(_construct_classical_mutation_classes(n))
    if types in possible_types[1:]:
        data.update(_construct_exceptional_mutation_classes(n))

    from sage.env import DOT_SAGE
    types_path = Path(DOT_SAGE) / 'cluster_algebra_quiver'
    types_path.mkdir(exist_ok=True)
    types_file = types_path / f'mutation_classes_{n}.dig6'
    from sage.misc.temporary_file import atomic_write
    with atomic_write(types_file, binary=True) as f:
        pickle.dump(data, f)
    if verbose:
        keys = sorted(data, key=str)
        print("\nThe following types are saved to file", types_file, "and will now be used to determine quiver mutation types:")
        print(keys)


def save_quiver_data(n, up_to=True, types='ClassicalExceptional', verbose=True):
    r"""
    Save mutation classes of certain quivers of ranks up to and equal
    to `n` or equal to `n` to
    ``DOT_SAGE/cluster_algebra_quiver/mutation_classes_n.dig6``.

    This data will then be used to determine quiver mutation types.

    INPUT:

    - ``n`` -- the rank (or the upper limit on the rank) of the mutation
      classes that are being saved

    - ``up_to`` -- (default: ``True``) if ``True``, saves data for
      ranks smaller than or equal to `n`; if ``False``, saves data
      for rank exactly `n`

    - ``types`` -- (default: ``'ClassicalExceptional'``) if all, saves data
      for both exceptional mutation-finite quivers and for classical
      quiver; the input 'Exceptional' or 'Classical' is also allowed
      to save only part of this data

    TESTS::

        sage: # needs sage.modules
        sage: from sage.combinat.cluster_algebra_quiver.quiver_mutation_type import save_quiver_data
        sage: save_quiver_data(2)
        <BLANKLINE>
        The following types are saved to file ... and will now be used to determine quiver mutation types:
        [('A', 1)]
        <BLANKLINE>
        The following types are saved to file ... and will now be used to determine quiver mutation types:
        [('A', (1, 1), 1), ('A', 2), ('B', 2), ('BC', 1, 1), ('G', 2)]
        sage: save_quiver_data(2,up_to=False)
        <BLANKLINE>
        The following types are saved to file ... and will now be used to determine quiver mutation types:
        [('A', (1, 1), 1), ('A', 2), ('B', 2), ('BC', 1, 1), ('G', 2)]
        sage: save_quiver_data(2,up_to=False, types='Classical')
        <BLANKLINE>
        The following types are saved to file ... and will now be used to determine quiver mutation types:
        [('A', (1, 1), 1), ('A', 2), ('B', 2), ('BC', 1, 1)]
        sage: save_quiver_data(2,up_to=False, types='Exceptional')
        <BLANKLINE>
        The following types are saved to file ... and will now be used to determine quiver mutation types:
        [('G', 2)]
        sage: save_quiver_data(2,up_to=False, verbose=False)
    """
    from sage.combinat.cluster_algebra_quiver.mutation_type import load_data
    if up_to is True:
        ranks = range(1, n + 1)
    elif up_to is False:
        ranks = [n]
    for i in ranks:
        _save_data_dig6(i, types=types, verbose=verbose)
    # we finally clear the load_data
    load_data.clear_cache()


def _bipartite_graph_to_digraph(g) -> DiGraph:
    """
    Return a digraph obtained from a bipartite graph ``g`` by choosing one
    set of the bipartition to be the set of sinks and the other to be the
    set of sources.

    EXAMPLES::

        sage: from sage.combinat.cluster_algebra_quiver.quiver_mutation_type import _bipartite_graph_to_digraph
        sage: G = Graph([(1, 2)])
        sage: _bipartite_graph_to_digraph(G)
        Digraph on 2 vertices
    """
    sources = g.bipartite_sets()[0]
    dg = DiGraph()
    dg.add_vertices(g)
    for a, b, c in g.edge_iterator():
        if a in sources:
            dg.add_edge(a, b, c)
        else:
            dg.add_edge(b, a, c)
    return dg


def _is_mutation_type(data) -> bool:
    """
    Return ``True`` if ``data`` is a QuiverMutationType.

    EXAMPLES::

        sage: from sage.combinat.cluster_algebra_quiver.quiver_mutation_type import _is_mutation_type
        sage: _is_mutation_type ( [ 'A', 2 ] )
        True
        sage: _is_mutation_type ( [ 'P', 1 ] )
        False
    """
    try:
        QuiverMutationType(data)
        return True
    except Exception:
        return False


def _mutation_type_error(data):
    r"""
    Output an error message because data which is not a valid quiver mutation
    type has been passed to QuiverMutationType.

    EXAMPLES::

        sage: QuiverMutationType( 'Christian', 'Stump' )  # indirect doctest
        Traceback (most recent call last):
        ...
        ValueError: ['Christian', 'Stump'] is not a valid quiver mutation type
                Finite types have the form [ '?', n ] for type ? and rank n
                Affine type A has the form [ 'A', [ i, j ], 1 ] for rank i+j
                Affine type ? has the form [ '?', k, \pm 1 ] for rank k+1
                Elliptic type ? has the form [ '?', k, [i, j] ] (1 <= i,j <= 3) for rank k+2
                For correct syntax in other types, please consult the documentation.
    """
    if data[2] is None:
        del data[2]
    return_str = str(data) + ' is not a valid quiver mutation type'
    return_str += '\n            Finite types have the form [ \'?\', n ] for type ? and rank n'
    return_str += '\n            Affine type A has the form [ \'A\', [ i, j ], 1 ] for rank i+j'
    return_str += '\n            Affine type ? has the form [ \'?\', k, \\pm 1 ] for rank k+1'
    return_str += '\n            Elliptic type ? has the form [ \'?\', k, [i, j] ] (1 <= i,j <= 3) for rank k+2'
    return_str += '\n            For correct syntax in other types, please consult the documentation.'

    raise ValueError(return_str)


def _edge_list_to_matrix(edges, nlist, mlist) -> matrix:
    r"""
    Return the matrix obtained from the edge list of a quiver.

    INPUT:

    - ``edges`` -- the list of edges
    - ``nlist`` -- the list of mutable vertices of the quiver
    - ``mlist`` -- the list of frozen vertices of the quiver

    OUTPUT:

    An `(n+m) \times n` matrix corresponding to the edge-list.

    EXAMPLES::

        sage: from sage.combinat.cluster_algebra_quiver.quiver_mutation_type import _edge_list_to_matrix
        sage: G = QuiverMutationType(['A', 2])._digraph
        sage: _edge_list_to_matrix(G.edges(sort=True), [0, 1], [])                      # needs sage.modules
        [ 0  1]
        [-1  0]

        sage: G2 = DiGraph([('a', 'b', 1)])
        sage: _edge_list_to_matrix(G2.edges(sort=True), ['a', 'b'], [])                 # needs sage.modules
        [ 0  1]
        [-1  0]

        sage: G3 = DiGraph([('a', 'b', 1), ('b', 'c', 1)])
        sage: _edge_list_to_matrix(G3.edges(sort=True), ['a', 'b'], ['c'])              # needs sage.modules
        [ 0  1]
        [-1  0]
        [ 0 -1]
    """
    n = len(nlist)
    nmlist = nlist + mlist
    nm = len(nmlist)
    M = matrix(ZZ, nm, n, sparse=True)
    for v1, v2, label in edges:
        if label is None:
            a, b = 1, -1
        elif label in ZZ:
            a, b = label, -label
        else:
            a, b = label
        if v1 in nlist:
            M[nmlist.index(v2), nmlist.index(v1)] = b
        if v2 in nlist:
            M[nmlist.index(v1), nmlist.index(v2)] = a
    return M



# sage_setup: distribution = sagemath-graphs
# cython: binding=True
r"""
Line graphs

This module gather everything which is related to line graphs. Right now, this
amounts to the following functions :

.. csv-table::
    :class: contentstable
    :widths: 30, 70
    :delim: |

    :meth:`line_graph` | Return the line graph of a given graph
    :meth:`is_line_graph` | Check whether a graph is a line graph
    :meth:`root_graph` | Return the root graph corresponding to the given graph

Author:

- Nathann Cohen (01-2013), :meth:`root_graph` method and module documentation.
  Written while listening to Nina Simone *"I wish I knew how it would feel to be
  free"*. Crazy good song. And *"Prendre ta douleur"*, too.

- David Coudert (10-2018), use maximal cliques iterator in :meth:`root_graph`,
  and use :meth:`root_graph` instead of forbidden subgraph search in
  :meth:`is_line_graph` (:issue:`26444`).

Definition
-----------

Given a graph `G`, the *line graph* `L(G)` of `G` is the graph such that

.. MATH::

    V(L(G)) =& E(G)\\
    E(L(G)) =& \{(e,e'):\text{ and }e,e'\text{ have a common endpoint in }G\}\\

The definition is extended to directed graphs. In this situation, there is an
arc `(e,e')` in `L(G)` if the destination of `e` is the origin of `e'`.

For more information, see the :wikipedia:`Line_graph`.

Root graph
----------

A graph whose line graph is `LG` is called the *root graph* of `LG`. The root
graph of a (connected) graph is unique ([Whi1932]_, [Har1969]_), except when
`LG=K_3`, as both `L(K_3)` and `L(K_{1,3})` are equal to `K_3`.

Here is how we can *"see"* `G` by staring (very intently) at `LG` :

  A graph `LG` is the line graph of `G` if there exists a collection
  `(S_v)_{v\in G}` of subsets of `V(LG)` such that :

  * Every `S_v` is a complete subgraph of `LG`.

  * Every `v\in LG` belongs to exactly two sets of the family `(S_v)_{v\in G}`.

  * Any two sets of `(S_v)_{v\in G}` have at most one common elements

  * For any edge `(u,v)\in LG` there exists a set of `(S_v)_{v\in G}` containing
    both `u` and `v`.

  In this family, each set `S_v` represent a vertex of `G`, and contains "the
  set of edges incident to `v` in `G`". Two elements `S_v,S_{v'}` have a
  nonempty intersection whenever `vv'` is an edge of `G`.

  Hence, finding the root graph of `LG` is the job of finding this collection of
  sets.

In particular, what we know for sure is that a maximal clique `S` of size `2` or
`\geq 4` in `LG` corresponds to a vertex of degree `|S|` in `G`, whose incident
edges are the elements of `S` itself.

The main problem lies with maximal cliques of size 3, i.e. triangles. Those we
have to split into two categories, *even* and *odd* triangles :

  A triangle `\{e_1,e_2,e_3\}\subseteq V(LG)` is said to be an *odd* triangle if
  there exists a vertex `e\in V(G)` incident to exactly *one* or *all* of
  `\{e_1,e_2,e_3\}`, and it is said to be *even* otherwise.

The very good point of this definition is that an inclusionwise maximal clique
which is an odd triangle will always correspond to a vertex of degree 3 in `G`,
while an even triangle could result from either a vertex of degree 3 in `G` or a
triangle in `G`. And in order to build the root graph we obviously have to
decide *which*.

Beineke proves in [Bei1970]_ that the collection of sets we are looking for
can be easily found. Indeed it turns out that it is the union of :

#. The family of all maximal cliques of `LG` of size 2 or `\geq 4`, as well as
   all odd triangles.

#. The family of all pairs of adjacent vertices which appear in exactly *one*
   maximal clique which is an even triangle.

There are actually four special cases to which the decomposition above does not
apply, i.e. graphs containing an edge which belongs to exactly two even
triangles. We deal with those independently.

* The :meth:`Complete graph
  <sage.graphs.graph_generators.GraphGenerators.CompleteGraph>` `K_3`.

* The :meth:`Diamond graph
  <sage.graphs.graph_generators.GraphGenerators.DiamondGraph>` -- the line graph
  of `K_{1,3}` plus an edge.

* The :meth:`Wheel graph
  <sage.graphs.graph_generators.GraphGenerators.WheelGraph>` on `4+1` vertices
  -- the line graph of the :meth:`Diamond graph
  <sage.graphs.graph_generators.GraphGenerators.DiamondGraph>`

* The :meth:`Octahedron
  <sage.graphs.graph_generators.GraphGenerators.OctahedralGraph>` -- the line
  graph of `K_4`.

This decomposition turns out to be very easy to implement :-)

.. WARNING::

    Even though the root graph is *NOT UNIQUE* for the triangle, this method
    returns `K_{1,3}` (and not `K_3`) in this case. Pay *very close* attention
    to that, for this answer is not theoretically correct : there is no unique
    answer in this case, and we deal with it by returning one of the two
    possible answers.

Functions
---------
"""


def is_line_graph(g, certificate=False):
    r"""
    Check whether the graph `g` is a line graph.

    INPUT:

    - ``certificate`` -- boolean; whether to return a certificate along with
      the boolean result. Here is what happens when ``certificate = True``:

      - If the graph is not a line graph, the method returns a pair ``(b,
        subgraph)`` where ``b`` is ``False`` and ``subgraph`` is a subgraph
        isomorphic to one of the 9 forbidden induced subgraphs of a line graph.

      - If the graph is a line graph, the method returns a triple ``(b,R,isom)``
        where ``b`` is ``True``, ``R`` is a graph whose line graph is the graph
        given as input, and ``isom`` is a map associating an edge of ``R`` to
        each vertex of the graph.

    .. NOTE::

        This method wastes a bit of time when the input graph is not connected.
        If you have performance in mind, it is probably better to only feed it
        with connected graphs only.

    .. SEEALSO::

        - The :mod:`line_graph <sage.graphs.line_graph>` module.

        - :meth:`~sage.graphs.graph_generators.GraphGenerators.line_graph_forbidden_subgraphs`
          -- the forbidden subgraphs of a line graph.

        - :meth:`~sage.graphs.generic_graph.GenericGraph.line_graph`

    EXAMPLES:

    A complete graph is always the line graph of a star::

        sage: graphs.CompleteGraph(5).is_line_graph()
        True

    The Petersen Graph not being claw-free, it is not a line
    graph::

        sage: graphs.PetersenGraph().is_line_graph()
        False

    This is indeed the subgraph returned::

        sage: C = graphs.PetersenGraph().is_line_graph(certificate=True)[1]             # needs sage.modules
        sage: C.is_isomorphic(graphs.ClawGraph())                                       # needs sage.modules
        True

    The house graph is a line graph::

        sage: g = graphs.HouseGraph()
        sage: g.is_line_graph()
        True

    But what is the graph whose line graph is the house ?::

        sage: # needs sage.modules
        sage: is_line, R, isom = g.is_line_graph(certificate=True)
        sage: R.sparse6_string()
        ':DaHI~'
        sage: R.show()                                                                  # needs sage.plot
        sage: isom
        {0: (0, 1), 1: (0, 2), 2: (1, 3), 3: (2, 3), 4: (3, 4)}

    TESTS:

    Disconnected graphs::

        sage: g = 2 * graphs.CycleGraph(3)
        sage: gl = g.line_graph().relabel(inplace=False)
        sage: new_g = gl.is_line_graph(certificate=True)[1]                             # needs sage.modules
        sage: g.line_graph().is_isomorphic(gl)                                          # needs sage.modules
        True

    Verify that :issue:`29740` is fixed::

        sage: g = Graph('O{e[{}^~z`MDZBZBkXzE^')
        sage: g.is_line_graph()
        False
    """
    g._scream_if_not_simple()

    def get_certificate(gg):
        """
        Assuming that gg is not a line graph, return a forbidden induced
        subgraph.
        """
        from sage.graphs.graph_generators import graphs
        for fg in graphs.line_graph_forbidden_subgraphs():
            h = gg.subgraph_search(fg, induced=True)
            if h is not None:
                return h

    if g.is_connected():
        try:
            # To check whether g is a line graph, we try to construct its root
            # graph
            R, isom = root_graph(g)
            if certificate:
                return True, R, isom
            return True
        except ValueError as VE:
            if str(VE) == "this graph is not a line graph !":
                # g is not a line graph
                if certificate:
                    return False, get_certificate(g)
                return False
            raise VE

    # g is not connected, so we apply the above procedure to each connected
    # component
    from sage.graphs.graph import Graph
    R = Graph()
    for gg in g.connected_components_subgraphs():
        try:
            RR, isom = root_graph(gg)
            R += RR
        except ValueError as VE:
            if str(VE) == "this graph is not a line graph !":
                # gg is not a line graph
                if certificate:
                    return False, get_certificate(gg)
                return False
            raise VE

    if certificate:
        _, isom = g.is_isomorphic(R.line_graph(labels=False), certificate=True)
        return True, R, isom
    return True


def line_graph(g, labels=True):
    """
    Return the line graph of the (di)graph ``g``.

    INPUT:

    - ``labels`` -- boolean (default: ``True``); whether edge labels should be
      taken in consideration. If ``labels=True``, the vertices of the line graph
      will be triples ``(u,v,label)``, and pairs of vertices otherwise.

    The line graph of an undirected graph G is an undirected graph H such that
    the vertices of H are the edges of G and two vertices e and f of H are
    adjacent if e and f share a common vertex in G. In other words, an edge in H
    represents a path of length 2 in G.

    The line graph of a directed graph G is a directed graph H such that the
    vertices of H are the edges of G and two vertices e and f of H are adjacent
    if e and f share a common vertex in G and the terminal vertex of e is the
    initial vertex of f. In other words, an edge in H represents a (directed)
    path of length 2 in G.

    .. NOTE::

        As a :class:`Graph` object only accepts hashable objects as vertices
        (and as the vertices of the line graph are the edges of the graph), this
        code will fail if edge labels are not hashable. You can also set the
        argument ``labels=False`` to ignore labels.

    .. SEEALSO::

        - The :mod:`line_graph <sage.graphs.line_graph>` module

        - :meth:`~sage.graphs.graph_generators.GraphGenerators.line_graph_forbidden_subgraphs`
          -- the forbidden subgraphs of a line graph

        - :meth:`~Graph.is_line_graph` -- tests whether a graph is a line graph

    EXAMPLES::

        sage: g = graphs.CompleteGraph(4)
        sage: h = g.line_graph()
        sage: h.vertices(sort=True)
        [(0, 1, None),
         (0, 2, None),
         (0, 3, None),
         (1, 2, None),
         (1, 3, None),
         (2, 3, None)]
        sage: h.am()                                                                    # needs sage.modules
        [0 1 1 1 1 0]
        [1 0 1 1 0 1]
        [1 1 0 0 1 1]
        [1 1 0 0 1 1]
        [1 0 1 1 0 1]
        [0 1 1 1 1 0]
        sage: h2 = g.line_graph(labels=False)
        sage: h2.vertices(sort=True)
        [(0, 1), (0, 2), (0, 3), (1, 2), (1, 3), (2, 3)]
        sage: h2.am() == h.am()                                                         # needs sage.modules
        True
        sage: g = DiGraph([[1..4], lambda i,j: i < j])
        sage: h = g.line_graph()
        sage: h.vertices(sort=True)
        [(1, 2, None),
         (1, 3, None),
         (1, 4, None),
         (2, 3, None),
         (2, 4, None),
         (3, 4, None)]
        sage: h.edges(sort=True)
        [((1, 2, None), (2, 3, None), None),
         ((1, 2, None), (2, 4, None), None),
         ((1, 3, None), (3, 4, None), None),
         ((2, 3, None), (3, 4, None), None)]

    TESTS:

    :issue:`13787`::

        sage: g = graphs.KneserGraph(7,1)
        sage: C = graphs.CompleteGraph(7)
        sage: C.is_isomorphic(g)
        True
        sage: C.line_graph().is_isomorphic(g.line_graph())
        True
    """
    cdef dict conflicts = {}
    cdef list elist = []

    g._scream_if_not_simple()
    if g._directed:
        from sage.graphs.digraph import DiGraph
        G = DiGraph()
        G.add_vertices(g.edge_iterator(labels=labels))
        for v in g:
            # Connect appropriate incident edges of the vertex v
            G.add_edges((e, f) for e in g.incoming_edge_iterator(v, labels=labels)
                        for f in g.outgoing_edge_iterator(v, labels=labels))
        return G

    from sage.graphs.graph import Graph
    G = Graph()

    # We must sort the edges' endpoints so that (1,2,None) is seen as the
    # same edge as (2,1,None).
    #
    # We do so by comparing hashes, just in case all the natural order (<)
    # on vertices would not be a total order (for instance when vertices are
    # sets). If two adjacent vertices have the same hash, then we store the
    # pair in the dictionary of conflicts

    # 1) List of vertices in the line graph
    for e in g.edge_iterator(labels=labels):
        if hash(e[0]) < hash(e[1]):
            elist.append(e)
        elif hash(e[0]) > hash(e[1]):
            elist.append((e[1], e[0]) + e[2:])
        else:
            # Settle the conflict arbitrarily
            conflicts[e] = e
            conflicts[(e[1], e[0]) + e[2:]] = e
            elist.append(e)

    G.add_vertices(elist)

    # 2) adjacencies in the line graph
    for v in g:
        elist = []

        # Add the edge to the list, according to hashes, as previously
        for e in g.edge_iterator(v, labels=labels):
            if hash(e[0]) < hash(e[1]):
                elist.append(e)
            elif hash(e[0]) > hash(e[1]):
                elist.append((e[1], e[0]) + e[2:])
            else:
                elist.append(conflicts[e])

        # All pairs of elements in elist are edges of the line graph
        while elist:
            x = elist.pop()
            for y in elist:
                G.add_edge(x, y)

    return G


def root_graph(g, verbose=False):
    r"""
    Return the root graph corresponding to the given graph ``g``.

    See the documentation of :mod:`sage.graphs.line_graph` to know how it works.

    INPUT:

    - ``g`` -- a graph

    - ``verbose`` -- boolean (default: ``False``); display some information
      about what is happening inside of the algorithm

    .. WARNING::

        This code assumes that `g` is a line graph, and is a connected,
        undirected graph without multiple edges.

    TESTS:

    All connected graphs on 6 vertices::

        sage: from sage.graphs.line_graph import root_graph
        sage: def test(g):
        ....:    gl = g.line_graph(labels=False)
        ....:    d = root_graph(gl)
        sage: for i,g in enumerate(graphs(6)): # long time
        ....:   if not g.is_connected():
        ....:     continue
        ....:   test(g)

    Non line-graphs::

        sage: root_graph(graphs.PetersenGraph())
        Traceback (most recent call last):
        ...
        ValueError: this graph is not a line graph !

        sage: root_graph(Graph('O{e[{}^~z`MDZBZBkXzE^'))
        Traceback (most recent call last):
        ...
        ValueError: this graph is not a line graph !

    Small corner-cases::

        sage: from sage.graphs.line_graph import root_graph
        sage: root_graph(graphs.CompleteGraph(3))
        (Complete bipartite graph of order 1+3: Graph on 4 vertices,
         {0: (0, 1), 1: (0, 2), 2: (0, 3)})
        sage: G, D = root_graph(graphs.OctahedralGraph()); G
        Complete graph: Graph on 4 vertices
        sage: G, D = root_graph(graphs.DiamondGraph()); G
        Graph on 4 vertices
        sage: G, D = root_graph(graphs.WheelGraph(5)); G
        Diamond Graph: Graph on 4 vertices
    """
    from sage.graphs.digraph import DiGraph

    if isinstance(g, DiGraph):
        raise ValueError("g cannot be a DiGraph !")
    if g.has_multiple_edges():
        raise ValueError("g cannot have multiple edges !")
    if not g.is_connected():
        raise ValueError("g is not connected !")
    # is_line_graph expects a particular error message when g is not a line graph
    not_line_graph = "this graph is not a line graph !"

    # Complete Graph ?
    if g.is_clique():
        from sage.graphs.generators.basic import CompleteBipartiteGraph
        return (CompleteBipartiteGraph(1, g.order()),
                {v: (0, 1 + i) for i, v in enumerate(g)})

    # Diamond Graph ?
    elif g.order() == 4 and g.size() == 5:
        from sage.graphs.graph import Graph
        root = Graph([(0, 1), (1, 2), (2, 0), (0, 3)])
        return (root,
                g.is_isomorphic(root.line_graph(labels=False), certificate=True)[1])

    # Wheel on 5 vertices ?
    elif g.order() == 5 and g.size() == 8 and min(g.degree()) == 3:
        from sage.graphs.generators.basic import DiamondGraph
        root = DiamondGraph()
        return (root,
                g.is_isomorphic(root.line_graph(labels=False), certificate=True)[1])

    # Octahedron ?
    elif g.order() == 6 and g.size() == 12 and g.is_regular(k=4):
        from sage.graphs.generators.platonic_solids import OctahedralGraph
        if g.is_isomorphic(OctahedralGraph()):
            from sage.graphs.generators.basic import CompleteGraph
            root = CompleteGraph(4)
            return (root,
                    g.is_isomorphic(root.line_graph(labels=False), certificate=True)[1])

    # From now on we can assume (thanks to Beineke) that no edge belongs to two
    # even triangles at once.

    # Better to work on integers... Everything takes more time otherwise.
    G = g.relabel(inplace=False)

    # Dictionary of (pairs of) cliques, i.e. the two cliques associated with
    # each vertex.
    cdef dict v_cliques = {v: [] for v in G}
    # All the even triangles we meet
    cdef list even_triangles = []

    # We iterate over all maximal cliques of the graph.

    # As method G.cliques_maximal() returns the list of all maximal cliques
    # (which takes an exponential time on general graphs, while it is obviously
    # polynomial on line graphs), we instead use IndependentSet to iterate over
    # all the maximal cliques. This way, we can say early that the graph is not
    # a line graph.

    from sage.graphs.independent_sets import IndependentSets
    for S in IndependentSets(G, maximal=True, complement=True):

        # Triangles... even or odd ?
        if len(S) == 3:

            # If a vertex of G has an odd number of neighbors among the vertices
            # of S, then the triangle is odd. We compute the list of such
            # vertices by taking the symmetric difference of the neighborhood of
            # our three vertices.
            #
            # Note that the elements of S do not appear in this set as they are
            # all seen exactly twice.

            odd_neighbors = set(G.neighbor_iterator(S[0]))
            odd_neighbors.symmetric_difference_update(G.neighbor_iterator(S[1]))
            odd_neighbors.symmetric_difference_update(G.neighbor_iterator(S[2]))

            # Even triangles
            if not odd_neighbors:
                even_triangles.append(tuple(S))
                continue

            # We manage odd triangles the same way we manage other cliques ...

        # We now associate the clique to all the vertices it contains.
        for v in S:
            if len(v_cliques[v]) == 2:
                raise ValueError(not_line_graph)
            v_cliques[v].append(tuple(S))

        if verbose:
            print("Added clique", S)

    # Deal with even triangles
    for u, v, w in even_triangles:

        # According to Beineke, we must go through all even triangles, and for
        # each triangle uvw consider its three pairs of adjacent vertices uv,
        # vw, wu. For all pairs xy among those such that xy do not appear
        # together in any clique we have found so far, we add xy to the list of
        # cliques describing our covering.

        for x, y in [(u, v), (v, w), (w, u)]:

            # If edge xy does not appear in any of the cliques associated with y
            if all(x not in C for C in v_cliques[y]):
                if len(v_cliques[y]) >= 2 or len(v_cliques[x]) >= 2:
                    raise ValueError("this graph is not a line graph !")

                v_cliques[x].append((x, y))
                v_cliques[y].append((x, y))

                if verbose:
                    print("Adding pair", (x, y),
                          "appearing in the even triangle", (u, v, w))

    # Deal with vertices contained in only one clique. All edges must be defined
    # by TWO endpoints, so we add a fake clique.
    for x, clique_list in v_cliques.items():
        if len(clique_list) == 1:
            clique_list.append((x,))

    # We now have all our cliques. Let's build the root graph to check that it
    # all fits !
    from sage.graphs.graph import Graph
    R = Graph()

    # Associates an integer to each clique
    cdef dict relabel = {}

    # Associates to each vertex of G its pair of coordinates in R
    cdef dict vertex_to_map = {}

    for v, L in v_cliques.items():

        # Add cliques to relabel dictionary
        for S in L:
            if S not in relabel:
                relabel[S] = len(relabel)

        # The coordinates of edge v
        vertex_to_map[v] = relabel[L[0]], relabel[L[1]]

    if verbose:
        print("Final associations :")
        for v, L in v_cliques.items():
            print(v, L)

    # We now build R
    R.add_edges(vertex_to_map.values())

    # If g is a line graph, then it is isomorphic to the line graph of the graph
    # R that we have constructed, so we return R (and the isomorphism).
    is_isom, isom = g.is_isomorphic(R.line_graph(labels=False), certificate=True)
    if is_isom:
        return R, isom
    raise ValueError(not_line_graph)



# sage_setup: distribution = sagemath-graphs
# sage.doctest: needs sage.rings.finite_rings sage.schemes
# cython: cdivision=True
r"""
Orthogonal arrays (find recursive constructions)

This module implements several functions to find recursive constructions of
:mod:`Orthogonal Arrays <sage.combinat.designs.orthogonal_arrays>`.

The main function of this module, i.e. :func:`find_recursive_construction`,
queries all implemented recursive constructions of designs implemented in
:mod:`~sage.combinat.designs.orthogonal_arrays_build_recursive` in order to
obtain an `OA(k,n)`.

:func:`find_recursive_construction` is called by the
:func:`~sage.combinat.designs.orthogonal_arrays.orthogonal_array` function.

.. csv-table::
    :class: contentstable
    :widths: 30, 70
    :delim: |

    :func:`find_recursive_construction` | Find a recursive construction of an `OA(k,n)` (calls all others ``find_*`` functions)
    :func:`find_product_decomposition` | Find `n_1n_2=n` to obtain an `OA(k,n)` by the product construction
    :func:`find_wilson_decomposition_with_one_truncated_group` | Find `rm+u=n` to obtain an `OA(k,n)` by Wilson's construction with one truncated column.
    :func:`find_wilson_decomposition_with_two_truncated_groups` | Find `rm+r_1+r_2=n` to obtain an `OA(k,n)` by Wilson's construction with two truncated columns.
    :func:`find_construction_3_3` | Find a decomposition for construction 3.3 from [AC07]_.
    :func:`find_construction_3_4` | Find a decomposition for construction 3.4 from [AC07]_.
    :func:`find_construction_3_5` | Find a decomposition for construction 3.5 from [AC07]_.
    :func:`find_construction_3_6` | Find a decomposition for construction 3.6 from [AC07]_.
    :func:`find_q_x` | Find integers `q,x` such that the `q-x` construction yields an `OA(k,n)`.
    :func:`find_thwart_lemma_3_5` | Find the values on which Lemma 3.5 from [Thwarts]_ applies.
    :func:`find_thwart_lemma_4_1` | Find a decomposition for Lemma 4.1 from [Thwarts]_.
    :func:`find_three_factor_product` | Find `n_1n_2n_3=n` to obtain an `OA(k,n)` by the three-factor product from [DukesLing14]_
    :func:`find_brouwer_separable_design` | Find `t(q^2+q+1)+x=n` to obtain an `OA(k,n)` by Brouwer's separable design construction.
    :func:`find_brouwer_van_rees_with_one_truncated_column` | Find `rm+x_1+...+x_c=n` such that the Brouwer-van Rees constructions yields a `OA(k,n)`.

REFERENCES:

.. [AC07] Concerning eight mutually orthogonal latin squares
  Julian R. Abel, Nicholas Cavenagh
  Journal of Combinatorial Designs
  Vol. 15, n.3, pp. 255-261
  2007

Functions
---------
"""

from sage.misc.cachefunc import cached_function
from sage.combinat.designs.orthogonal_arrays import orthogonal_array
from sage.rings.integer cimport smallInteger
from sage.arith.misc import prime_powers


@cached_function
def find_recursive_construction(k, n):
    r"""
    Find a recursive construction of an `OA(k,n)` (calls all others ``find_*`` functions).

    This determines whether an `OA(k,n)` can be built through the following
    constructions:

    - :func:`~sage.combinat.designs.orthogonal_arrays.wilson_construction`
    - :func:`~sage.combinat.designs.orthogonal_arrays_build_recursive.construction_3_3`
    - :func:`~sage.combinat.designs.orthogonal_arrays_build_recursive.construction_3_4`
    - :func:`~sage.combinat.designs.orthogonal_arrays_build_recursive.construction_3_5`
    - :func:`~sage.combinat.designs.orthogonal_arrays_build_recursive.construction_3_6`
    - :func:`~sage.combinat.designs.orthogonal_arrays_build_recursive.construction_q_x`
    - :func:`~sage.combinat.designs.orthogonal_arrays_build_recursive.thwart_lemma_3_5`
    - :func:`~sage.combinat.designs.orthogonal_arrays_build_recursive.thwart_lemma_4_1`
    - :func:`~sage.combinat.designs.orthogonal_arrays_build_recursive.three_factor_product`
    - :func:`~sage.combinat.designs.orthogonal_arrays_build_recursive.brouwer_separable_design`

    INPUT:

    - ``k``, ``n`` -- integers

    OUTPUT:

    Return a pair ``f,args`` such that ``f(*args)`` returns the requested `OA`
    if possible, and ``False`` otherwise.

    EXAMPLES::

        sage: from sage.combinat.designs.orthogonal_arrays_find_recursive import find_recursive_construction
        sage: from sage.combinat.designs.orthogonal_arrays import is_orthogonal_array
        sage: count = 0
        sage: for n in range(10,150):
        ....:     k = designs.orthogonal_arrays.largest_available_k(n)
        ....:     if find_recursive_construction(k,n):
        ....:         count = count + 1
        ....:         f,args = find_recursive_construction(k,n)
        ....:         OA = f(*args)
        ....:         assert is_orthogonal_array(OA,k,n,2,verbose=True)
        sage: count
        56
    """
    assert k > 3

    for find_c in [find_product_decomposition,
                   find_wilson_decomposition_with_one_truncated_group,
                   find_wilson_decomposition_with_two_truncated_groups,
                   find_construction_3_3,
                   find_construction_3_4,
                   find_construction_3_5,
                   find_construction_3_6,
                   find_q_x,
                   find_thwart_lemma_3_5,
                   find_thwart_lemma_4_1,
                   find_three_factor_product,
                   find_brouwer_separable_design,
                   find_brouwer_van_rees_with_one_truncated_column]:
        res = find_c(k,n)
        if res:
            return res
    return False


cpdef find_product_decomposition(int k, int n):
    r"""
    Find `n_1n_2=n` to obtain an `OA(k,n)` by the product construction.

    If Sage can build a `OA(k,n_1)` and a `OA(k,n_2)` such that `n=n_1\times
    n_2` then a `OA(k,n)` can be built by a product construction (which
    correspond to Wilson's construction with no truncated column). This
    function look for a pair of integers `(n_1,n_2)` with `n1 \leq n_2`, `n_1
    \times n_2 = n` and such that both an `OA(k,n_1)` and an `OA(k,n_2)` are
    available.

    INPUT:

    - ``k``, ``n`` -- integers

    OUTPUT:

    A pair ``f,args`` such that ``f(*args)`` is an `OA(k,n)` or ``False`` if no
    product decomposition was found.

    EXAMPLES::

        sage: from sage.combinat.designs.orthogonal_arrays_find_recursive import find_product_decomposition
        sage: f,args = find_product_decomposition(6, 84)
        sage: args
        (None, 6, 7, 12, (), False)
        sage: _ = f(*args)
    """
    cdef int n1,n2
    for n1 in range(2, n):
        n2 = n/n1  # n2 is decreasing along the loop
        if n2 < n1:
            break
        if n%n1:  # we want to iterate only through divisors of n1... it seems
                  # faster to use that rather than calling the divisors function
            continue
        if is_available(k, n1) and is_available(k, n2):
            from sage.combinat.designs.orthogonal_arrays import wilson_construction
            return wilson_construction, (None,k,n1,n2,(),False)
    return False

cpdef find_wilson_decomposition_with_one_truncated_group(int k, int n):
    r"""
    Find `rm+u=n` to obtain an `OA(k,n)` by Wilson's construction with one truncated column.

    This function looks for possible integers `m,t,u` satisfying that `mt+u=n` and
    such that Sage knows how to build a `OA(k,m)`, `OA(k,m+1)`, `OA(k+1,t)` and a
    `OA(k,u)`.

    INPUT:

    - ``k``, ``n`` -- integers

    OUTPUT:

    A pair ``f,args`` such that ``f(*args)`` is an `OA(k,n)` or ``False`` if no
    decomposition with one truncated block was found.

    EXAMPLES::

        sage: from sage.combinat.designs.orthogonal_arrays_find_recursive import find_wilson_decomposition_with_one_truncated_group
        sage: f,args = find_wilson_decomposition_with_one_truncated_group(4,38)
        sage: args
        (None, 4, 5, 7, (3,), False)
        sage: _ = f(*args)

        sage: find_wilson_decomposition_with_one_truncated_group(4,20)
        False
    """
    cdef int r,u,m
    # If there exists a TD(k+1,t) then k+1 < t+2, i.e. k <= t
    for r in range(max(1,k),n-1):
        u = n%r
        # We ensure that 1<=u, and that there can exists a TD(k,u), i.e k<u+2
        # (unless u == 1)
        if u == 0 or (u>1 and k >= u+2):
            continue

        m = n // r
        # If there exists a TD(k,m) then k<m+2
        if k >= m+2:
            break

        if (is_available(k, m) and is_available(k, m + 1) and
            is_available(k + 1, r) and is_available(k, u)):
            from sage.combinat.designs.orthogonal_arrays import wilson_construction
            return wilson_construction, (None,k,r,m,(u,),False)

    return False

cpdef find_wilson_decomposition_with_two_truncated_groups(int k, int n):
    r"""
    Find `rm+r_1+r_2=n` to obtain an `OA(k,n)` by Wilson's construction with two truncated columns.

    Look for integers `r,m,r_1,r_2` satisfying `n=rm+r_1+r_2` and `1\leq r_1,r_2<r`
    and such that the following designs exist : `OA(k+2,r)`, `OA(k,r1)`,
    `OA(k,r2)`, `OA(k,m)`, `OA(k,m+1)`, `OA(k,m+2)`.

    INPUT:

    - ``k``, ``n`` -- integers

    OUTPUT:

    A pair ``f,args`` such that ``f(*args)`` is an `OA(k,n)` or ``False`` if no
    decomposition with two truncated blocks was found.

    EXAMPLES::

        sage: from sage.combinat.designs.orthogonal_arrays_find_recursive import find_wilson_decomposition_with_two_truncated_groups
        sage: f,args = find_wilson_decomposition_with_two_truncated_groups(5,58)
        sage: args
        (None, 5, 7, 7, (4, 5), False)
        sage: _ = f(*args)
    """
    cdef int r,m_min,m_max,m,r1_min,r1_max,r1,r2,r1_p_r2
    for r in [1] + list(range(k+1, n-2)):
        # as r*1+1+1 <= n and because we need
        # an OA(k+2,r), necessarily r=1 or r >= k+1
        if not is_available(k+2,r):
            continue
        m_min = (n - (2*r-2))/r
        m_max = (n - 2)/r
        if m_min > 1:
            m_values = list(range(max(m_min, k - 1), m_max + 1))
        else:
            m_values = [1] + list(range(k - 1, m_max + 1))
        for m in m_values:
            r1_p_r2 = n-r*m # the sum of r1+r2
                            # it is automatically >= 2 since m <= m_max
            if (r1_p_r2 > 2*r-2 or
                not is_available(k,m  ) or
                not is_available(k,m+1) or
                not is_available(k,m+2)):
                continue

            r1_min = r1_p_r2 - (r-1)
            r1_max = min(r-1, r1_p_r2)
            if r1_min > 1:
                r1_values = range(max(k - 1, r1_min), r1_max + 1)
            else:
                r1_values = [1] + list(range(k-1, r1_max + 1))
            for r1 in r1_values:
                if not is_available(k,r1):
                    continue
                r2 = r1_p_r2-r1
                if is_available(k,r2):
                    assert n == r*m+r1+r2
                    from sage.combinat.designs.orthogonal_arrays import wilson_construction
                    return wilson_construction, (None,k,r,m,(r1,r2),False)
    return False

cpdef find_construction_3_3(int k, int n):
    r"""
    Find a decomposition for construction 3.3 from [AC07]_.

    INPUT:

    - ``k``, ``n`` -- integers

    .. SEEALSO::

        :func:`~sage.combinat.designs.orthogonal_arrays_build_recursive.construction_3_3`

    OUTPUT:

    A pair ``f,args`` such that ``f(*args)`` returns the requested OA.

    EXAMPLES::

        sage: from sage.combinat.designs.orthogonal_arrays_find_recursive import find_construction_3_3
        sage: find_construction_3_3(11,177)[1]
        (11, 11, 16, 1)
        sage: find_construction_3_3(12,11)
    """
    cdef int mm,nn,i
    for mm in range(k-1, n//2+1):
        if not(is_available(k, mm) and is_available(k, mm + 1)):
            continue

        for nn in range(2, n//mm+1):
            i = n-nn*mm
            if i <= 0:
                continue

            if is_available(k + i, nn) and is_available(k, mm + i):
                from sage.combinat.designs.orthogonal_arrays_build_recursive import construction_3_3
                return construction_3_3, (k, nn, mm, i)

cpdef find_construction_3_4(int k, int n):
    r"""
    Find a decomposition for construction 3.4 from [AC07]_.

    INPUT:

    - ``k``, ``n`` -- integers

    .. SEEALSO::

        :func:`~sage.combinat.designs.orthogonal_arrays_build_recursive.construction_3_4`

    OUTPUT:

    A pair ``f,args`` such that ``f(*args)`` returns the requested OA.

    EXAMPLES::

        sage: from sage.combinat.designs.orthogonal_arrays_find_recursive import find_construction_3_4
        sage: find_construction_3_4(8,196)[1]
        (8, 25, 7, 12, 9)
        sage: find_construction_3_4(9,24)
    """
    cdef int mm,nn,i,r,s
    for mm in range(k-1,n/2+1):
        if (not is_available(k,mm+0) or
            not is_available(k,mm+1) or
            not is_available(k,mm+2)):
            continue

        for nn in range(2, n//mm+1):
            i = n-nn*mm
            if i<=0:
                continue

            for s in range(1,min(i,nn)):
                r = i-s
                if (is_available(k + r + 1, nn) and
                    is_available(k, s) and
                    (is_available(k, mm + r) or is_available(k, mm + r + 1))):
                    from sage.combinat.designs.orthogonal_arrays_build_recursive import construction_3_4
                    return construction_3_4, (k, nn, mm, r, s)

cpdef find_construction_3_5(int k, int n):
    r"""
    Find a decomposition for construction 3.5 from [AC07]_.

    INPUT:

    - ``k``, ``n`` -- integers

    .. SEEALSO::

        :func:`~sage.combinat.designs.orthogonal_arrays_build_recursive.construction_3_5`

    OUTPUT:

    A pair ``f,args`` such that ``f(*args)`` returns the requested OA.

    EXAMPLES::

        sage: from sage.combinat.designs.orthogonal_arrays_find_recursive import find_construction_3_5
        sage: find_construction_3_5(8,111)[1]
        (8, 13, 6, 9, 11, 13)
        sage: find_construction_3_5(9,24)
    """
    cdef int mm,i,nn,r,s,t
    for mm in range(2, n//2+1):
        if (mm+3 >= n or
            not is_available(k,mm+1) or
            not is_available(k,mm+2) or
            not is_available(k,mm+3)):
            continue

        for nn in range(2, n//mm+1):
            i = n-nn*mm
            if i<=0:
                continue

            if not is_available(k+3,nn):
                continue

            # Enumerate all  r,s,t<nn such that r+s+t=i and r<=s
            for s in range(min(i+1,nn)):
                for r in range(max(0,i-nn-s), min(s+1,i-s+1,nn)):
                    t = i - r - s
                    if ((nn-r-1)*(nn-s) < t         and
                        (r==0 or is_available(k,r)) and
                        (s==0 or is_available(k,s)) and
                        (t==0 or is_available(k,t))):
                        from sage.combinat.designs.orthogonal_arrays_build_recursive import construction_3_5
                        return construction_3_5, (k,nn,mm,r,s,t)

cpdef find_construction_3_6(int k, int n):
    r"""
    Find a decomposition for construction 3.6 from [AC07]_.

    INPUT:

    - ``k``, ``n`` -- integers

    .. SEEALSO::

        :func:`~sage.combinat.designs.orthogonal_arrays_build_recursive.construction_3_6`

    OUTPUT:

    A pair ``f,args`` such that ``f(*args)`` returns the requested OA.

    EXAMPLES::

        sage: from sage.combinat.designs.orthogonal_arrays_find_recursive import find_construction_3_6
        sage: find_construction_3_6(8,95)[1]
        (8, 13, 7, 4)
        sage: find_construction_3_6(8,98)
    """
    cdef int mm,nn,i

    for mm in range(k-1,n/2+1):
        if (not is_available(k,mm+0) or
            not is_available(k,mm+1) or
            not is_available(k,mm+2)):
            continue

        for nn in range(2, n//mm+1):
            i = n-nn*mm
            if i<=0:
                continue

            if (is_available(k+i,nn) and
                smallInteger(nn).is_prime_power()):
                from sage.combinat.designs.orthogonal_arrays_build_recursive import construction_3_6
                return construction_3_6, (k,nn,mm,i)

cpdef find_q_x(int k, int n):
    r"""
    Find integers `q,x` such that the `q-x` construction yields an `OA(k,n)`.

    See the documentation of :func:`~sage.combinat.designs.orthogonal_arrays_build_recursive.construction_q_x` to find out what
    hypotheses the integers `q,x` must satisfy.

    .. WARNING::

        For efficiency reasons, this function checks that Sage can build an
        `OA(k+1,q-x-1)` and an `OA(k+1,q-x+1)`, which is stronger than checking
        that Sage can build a `OA(k,q-x-1)-(q-x-1).OA(k,1)` and a
        `OA(k,q-x+1)-(q-x+1).OA(k,1)`. The latter would trigger a lot of
        independent set computations in
        :func:`sage.combinat.designs.orthogonal_arrays.incomplete_orthogonal_array`.

    INPUT:

    - ``k``, ``n`` -- integers

    .. SEEALSO::

        :func:`~sage.combinat.designs.orthogonal_arrays_build_recursive.construction_q_x`

    EXAMPLES::

        sage: from sage.combinat.designs.orthogonal_arrays_find_recursive import find_q_x
        sage: find_q_x(10,9)
        False
        sage: find_q_x(9,158)[1]
        (9, 16, 6)
    """
    cdef int q,x

    # n = (q-1)*(q-x) + x + 2
    #   = q^2 - q*x - q + 2*x + 2
    for q in range(max(3,k+2),n):
        # n-q**2+q-2 = 2x-qx
        #            = x(2-q)
        x = (n-q**2+q-2)/(2-q)
        if (x < q and
            0 < x and
            n == (q-1)*(q-x)+x+2             and
            is_available(k+1,q-x-1)          and
            is_available(k+1,q-x+1)          and
            # The next is always True, because q is a prime power
            # is_available(k+1,q) and
            is_available(k, x+2 )            and
            smallInteger(q).is_prime_power()):
            from sage.combinat.designs.orthogonal_arrays_build_recursive import construction_q_x
            return construction_q_x, (k,q,x)
    return False

cpdef find_thwart_lemma_3_5(int k, int N):
    r"""
    Find the values on which Lemma 3.5 from [Thwarts]_ applies.

    OUTPUT:

    A pair ``(f,args)`` such that ``f(*args)`` returns an `OA(k,n)` or ``False``
    if the construction is not available.

    .. SEEALSO::

        :func:`~sage.combinat.designs.orthogonal_arrays_build_recursive.thwart_lemma_3_5`

    EXAMPLES::

        sage: from sage.combinat.designs.orthogonal_arrays_find_recursive import find_thwart_lemma_3_5
        sage: from sage.combinat.designs.designs_pyx import is_orthogonal_array

        sage: f,args = find_thwart_lemma_3_5(7,66)
        sage: args
        (7, 9, 7, 1, 1, 1, 0, False)
        sage: OA = f(*args)
        sage: is_orthogonal_array(OA,7,66,2)
        True

        sage: f,args = find_thwart_lemma_3_5(6,100)
        sage: args
        (6, 8, 10, 8, 7, 5, 0, True)
        sage: OA = f(*args)
        sage: is_orthogonal_array(OA,6,100,2)
        True

    Some values from [Thwarts]_::

        sage: kn = ((10,1046), (10,1048), (10,1059), (11,1524),
        ....:       (11,2164), (12,3362), (12,3992),  (12,3994))
        sage: for k,n in kn:
        ....:     print("{} {} {}".format(k,n,find_thwart_lemma_3_5(k,n)[1]))
        10 1046 (10, 13, 79, 9, 1, 0, 9, False)
        10 1048 (10, 13, 79, 9, 1, 0, 11, False)
        10 1059 (10, 13, 80, 9, 1, 0, 9, False)
        11 1524 (11, 19, 78, 16, 13, 13, 0, True)
        11 2164 (11, 27, 78, 23, 19, 16, 0, True)
        12 3362 (12, 16, 207, 13, 13, 11, 13, True)
        12 3992 (12, 19, 207, 16, 13, 11, 19, True)
        12 3994 (12, 19, 207, 16, 13, 13, 19, True)

        sage: for k,n in kn:                                                     # not tested -- too long
        ....:     assert designs.orthogonal_array(k,n,existence=True) is True
    """
    from sage.combinat.designs.orthogonal_arrays_build_recursive import thwart_lemma_3_5
    cdef int n,m,a,b,c,d,NN,na,nb,nc

    for n in prime_powers(k+2,N-2): # There must exist a OA(k+3,n) thus n>=k+2
                                    # At least 3 columns are nonempty thus n<N-2

        # we look for (m,n,a,b,c,d) with N = mn + a + b + c (+d) and
        # 0 <= a,b,c,d <= n
        # hence we have N/n-4 <= m <= N/n

        # 1. look for m,a,b,c,d with complement=False
        # (we restrict to a >= b >= c)
        for m in range(max(k-1,(N+n-1)/n-4), N/n+1):
            if not (is_available(k,m+0) and
                    is_available(k,m+1) and
                    is_available(k,m+2)):
                continue

            NN = N - n*m
            # as a >= b >= c and d <= n we can restrict the start of the loops
            for a in range(max(0, (NN-n+2)/3), min(n, NN)+1): # (NN-n+2)/3 <==> ceil((NN-n)/3)x
                if not is_available(k,a):
                    continue
                for b in range(max(0, (NN-n-a+1)/2), min(a, n+1-a, NN-a)+1):
                    if not is_available(k,b):
                        continue
                    for c in range(max(0, NN-n-a-b), min(b, n+1-a-b, NN-a-b)+1):
                        if not is_available(k,c):
                            continue

                        d = NN - (a + b + c)  # necessarily 0 <= d <= n
                        if d == 0:
                            return thwart_lemma_3_5, (k,n,m,a,b,c,0,False)
                        elif (k+4 <= n+1 and
                              is_available(k, d ) and
                              is_available(k,m+3)):
                            return thwart_lemma_3_5, (k,n,m,a,b,c,d,False)

        # 2. look for m,a,b,c,d with complement=True
        # (we restrict to a >= b >= c)
        for m in range(max(k-2,N/n-4), (N+n-1)/n):
            if not (is_available(k,m+1) and
                    is_available(k,m+2) and
                    is_available(k,m+3)):
                continue

            NN = N - n*m
            for a in range(max(0, (NN-n+2)/3), min(n, NN)+1):
                # (NN-n+2)/3 <==> ceil((NN-n)/3)
                if not is_available(k,a):
                    continue
                na = n-a
                for b in range(max(0, (NN-n-a+1)/2), min(a, NN-a)+1):
                    nb = n-b
                    if na+nb > n+1 or not is_available(k,b):
                        continue
                    for c in range(max(0, NN-n-a-b), min(b, NN-a-b)+1):
                        nc = n-c
                        if na+nb+nc > n+1 or not is_available(k,c):
                            continue

                        d = NN - (a + b + c)  # necessarily d <= n
                        if d == 0:
                            return thwart_lemma_3_5, (k,n,m,a,b,c,0,True)
                        elif (k+4 <= n+1 and
                              is_available(k, d ) and
                              is_available(k,m+4)):
                            return thwart_lemma_3_5, (k,n,m,a,b,c,d,True)

    return False

cpdef find_thwart_lemma_4_1(int k, int n):
    r"""
    Find a decomposition for Lemma 4.1 from [Thwarts]_.

    INPUT:

    - ``k``, ``n`` -- integers

    .. SEEALSO::

        :func:`~sage.combinat.designs.orthogonal_arrays_build_recursive.thwart_lemma_4_1`

    OUTPUT:

    A pair ``f,args`` such that ``f(*args)`` returns the requested OA.

    EXAMPLES::

        sage: from sage.combinat.designs.orthogonal_arrays_find_recursive import find_thwart_lemma_4_1
        sage: find_thwart_lemma_4_1(10,408)[1]
        (10, 13, 28)
        sage: find_thwart_lemma_4_1(10,50)
        False
    """
    cdef int p,i,imax,nn,mm

    #      n  = nn*mm+4(nn-2)
    # <=> n+8 = nn(mm+4)
    #
    # nn is a prime power dividing n+8
    for p,imax in smallInteger(n+8).factor():
        nn = 1
        for i in range(1,imax+1):
            nn *= p
            mm = (n+8)/nn-4
            if (k+4 > nn+1 or
                mm <= 1 or
                nn % 3 == 2 or
                not is_available(k,nn-2) or
                not is_available(k,mm+1) or
                not is_available(k,mm+3) or
                not is_available(k,mm+4)):
                continue

            from sage.combinat.designs.orthogonal_arrays_build_recursive import thwart_lemma_4_1
            return thwart_lemma_4_1,(k,nn,mm)

    return False

cpdef find_three_factor_product(int k, int n):
    r"""
    Find `n_1n_2n_3=n` to obtain an `OA(k,n)` by the three-factor product from [DukesLing14]_.

    INPUT:

    - ``k``, ``n`` -- integers

    .. SEEALSO::

        :func:`~sage.combinat.designs.orthogonal_arrays_build_recursive.three_factor_product`

    OUTPUT:

    A pair ``f,args`` such that ``f(*args)`` returns the requested OA.

    EXAMPLES::

        sage: from sage.combinat.designs.orthogonal_arrays_find_recursive import find_three_factor_product
        sage: find_three_factor_product(10,648)[1]
        (9, 8, 9, 9)
        sage: find_three_factor_product(10,50)
        False
    """
    cdef int n1,n2,n3

    # we want to write n=n1*n2*n3 where n1<=n2<=n3 and we can build:
    # - a OA(k-1,n1)
    # - a OA( k ,n2)
    # - a OA( k ,n3)
    for n1 in smallInteger(n).divisors()[1:-1]:
        if not is_available(k-1,n1):
            continue
        for n2 in smallInteger(n/n1).divisors():
            n3 = n/n1/n2
            if (n2<n1 or
                n3<n2 or
                not is_available(k,n2) or
                not is_available(k,n3)):
                continue
            from sage.combinat.designs.orthogonal_arrays_build_recursive import three_factor_product
            return three_factor_product,(k-1,n1,n2,n3)

    return False

cpdef find_brouwer_separable_design(int k, int n):
    r"""
    Find `t(q^2+q+1)+x=n` to obtain an `OA(k,n)` by Brouwer's separable design construction.

    INPUT:

    - ``k``, ``n`` -- integers

    The assumptions made on the parameters `t,q,x` are explained in the
    documentation of
    :func:`~sage.combinat.designs.orthogonal_arrays_build_recursive.brouwer_separable_design`.

    EXAMPLES::

        sage: from sage.combinat.designs.orthogonal_arrays_find_recursive import find_brouwer_separable_design
        sage: find_brouwer_separable_design(5,13)[1]
        (5, 1, 3, 0)
        sage: find_brouwer_separable_design(5,14)
        False
    """
    from sage.combinat.designs.orthogonal_arrays_build_recursive import brouwer_separable_design
    cdef int q, x, baer_subplane_size, max_t, min_t, t, e1, e2, e3, e4

    for q in prime_powers(2, n):
        baer_subplane_size = q**2 + q + 1
        if baer_subplane_size > n:
            break
        #                       x <= q^2+1
        # <=>        n-t(q^2+q+1) <= q^2+1
        # <=>             n-q^2-1 <= t(q^2+q+1)
        # <=> (n-q^2-1)/(q^2+q+1) <= t

        min_t = (n - q**2 - 1) / baer_subplane_size
        max_t = min(n / baer_subplane_size, q**2 - q + 1)

        for t in range(min_t, max_t + 1):
            x = n - t * baer_subplane_size
            e1 = int(x != q**2 - q - t)
            e2 = int(x != 1)
            e3 = int(x != q**2)
            e4 = int(x != t + q + 1)

            # i)
            if (x == 0 and
                    is_available(k, t) and is_available(k, t + q)):
                return brouwer_separable_design, (k, t, q, x)

            # ii)
            elif (x == t + q and
                  is_available(k + e3, t) and
                  is_available(k, t + q) and
                  is_available(k + 1, t + q + 1)):
                return brouwer_separable_design, (k, t, q, x)

            # iii)
            elif (x == q**2 - q + 1 - t and
                  is_available(k, x) and
                  is_available(k + e2, t + 1)
                  and is_available(k + 1, t + q)):
                return brouwer_separable_design, (k, t, q, x)

            # iv)
            elif (x == q**2 + 1 and
                  is_available(k, x) and
                  is_available(k + e4, t + 1) and
                  is_available(k + 1, t + q + 1)):
                return brouwer_separable_design, (k, t, q, x)

            # v)
            elif (0 < x < q**2 - q + 1 - t and (e1 or e2) and
                  is_available(k, x) and
                  is_available(k + e1, t) and
                  is_available(k + e2, t + 1) and
                  is_available(k + 1, t + q)):
                return brouwer_separable_design, (k, t, q, x)

            # vi)
            elif (t + q < x < q**2 + 1 and (e3 or e4) and
                  is_available(k, x) and
                  is_available(k + e3, t) and
                  is_available(k + e4, t + 1) and
                  is_available(k + 1, t + q + 1)):
                return brouwer_separable_design, (k, t, q, x)

    return False

# Associates to n the list of k,x with x>1 such that there exists an
# OA(k,n+x)-OA(k,x). Useful in find_brouwer_separable_design
from sage.combinat.designs.database import QDM as __QDM
cdef dict _QDM = __QDM
cdef dict ioa_indexed_by_n_minus_x = {}
for x in _QDM.itervalues():
    for (n, _, _, u), (k, _) in x.items():
        if u > 1:
            if n not in ioa_indexed_by_n_minus_x:
                ioa_indexed_by_n_minus_x[n] = []
            ioa_indexed_by_n_minus_x[n].append((k, u))


def int_as_sum(int value, list S, int k_max):
    r"""
    Return a tuple `(s_1, s_2, \ldots, s_k)` of less then `k_max` elements of `S` such
    that `value = s_1 + s_2 + \ldots + s_k`. If there is no such tuples then the
    function returns ``None``.

    INPUT:

    - ``value`` -- integer

    - ``S`` -- list of integers

    - ``k_max`` -- integer

    EXAMPLES::

        sage: from sage.combinat.designs.orthogonal_arrays_find_recursive import int_as_sum
        sage: D = int_as_sum(21,[5,12],100)
        sage: for k in range(20,40):
        ....:     print("{} {}".format(k, int_as_sum(k,[5,12],100)))
        20 (5, 5, 5, 5)
        21 None
        22 (12, 5, 5)
        23 None
        24 (12, 12)
        25 (5, 5, 5, 5, 5)
        26 None
        27 (12, 5, 5, 5)
        28 None
        29 (12, 12, 5)
        30 (5, 5, 5, 5, 5, 5)
        31 None
        32 (12, 5, 5, 5, 5)
        33 None
        34 (12, 12, 5, 5)
        35 (5, 5, 5, 5, 5, 5, 5)
        36 (12, 12, 12)
        37 (12, 5, 5, 5, 5, 5)
        38 None
        39 (12, 12, 5, 5, 5)
    """
    cdef int i,j,v,vv,max_value
    cdef dict D,new_D,last_D
    last_D = D = {value:tuple()}
    max_value = max(S)

    if k_max * max_value < value:
        return None

    # The answer for a given k can be easily deduced from the answer
    # for k-1. That's how we build the list, incrementally starting
    # from k=0
    for j in range(k-1,-1,-1):
        new_D = {}
        for i in S:
            for v in last_D:
                vv = v-i
                if vv == 0:
                    return D[v] + (i,)
                if (vv > 0            and   # The new integer i is too big
                    vv <= j*max_value and   # The new integer i is too small
                    vv not in D       and   # We had it in D     already
                    vv not in new_D):       # We had it in new_D already
                    new_D[vv] = D[v]+(i,)
        if not new_D:
            break
        D.update(new_D)
        last_D = new_D

    return None


cpdef find_brouwer_van_rees_with_one_truncated_column(int k, int n):
    r"""
    Find `rm+x_1+...+x_c=n` such that the Brouwer-van Rees constructions yields a `OA(k,n)`.

    Let `n=rm+\sum_{1\leq i\leq c}` such that `c\leq r`. The
    generalization of Wilson's construction found by Brouwer and van
    Rees (with one truncated column) ensures that an `OA(k,n)` exists
    if the following designs exist: `OA(k+1,r)`, `OA(k,m)`,
    `OA(k,\sum_{1\leq i\leq c} u_i)`, `OA(k,m+x_1)-OA(k,x_1)`, ...,
    `OA(k,m+x_c)-OA(k,x_c)`.

    For more information, see the documentation of
    :func:`~sage.combinat.designs.orthogonal_arrays.wilson_construction`.

    INPUT:

    - ``k``, ``n`` -- integers

    EXAMPLES::

        sage: from sage.combinat.designs.orthogonal_arrays_find_recursive import find_brouwer_van_rees_with_one_truncated_column
        sage: find_brouwer_van_rees_with_one_truncated_column(5,53)[1]
        (None, 5, 7, 7, [[(2, 1), (2, 1)]])
        sage: find_brouwer_van_rees_with_one_truncated_column(6,96)[1]
        (None, 6, 7, 13, [[(3, 1), (1, 1), (1, 1)]])
    """
    cdef list available_multipliers
    cdef int kk,uu,r,m,remainder,max_multiplier
    cdef tuple values

    # We write n=rm+remainder
    for m in range(2, n//2):
        if not is_available(k,m):
            continue

        # List of x such that a OA(k,m+x)-OA(k,x) exists
        #
        # This is the list of integers that can be used as multipliers
        # for the points of the truncated column
        available_multipliers = []
        if is_available(k,m+1):
            available_multipliers.append(1)
        for kk,uu in ioa_indexed_by_n_minus_x.get(m,[]):
            if kk>=k:
                available_multipliers.append(uu)

        # We stop if there is no multiplier, or if 1 is the only
        # multiplier (those cases are handled by other functions)
        if (not available_multipliers or
            (len(available_multipliers) == 1 and available_multipliers[0] == 1)):
            continue

        max_multiplier = max(available_multipliers)
        for r in range(2, n//m+1):
            remainder = n-r*m
            if (remainder > r*max_multiplier or
                not is_available(k+1,r) or
                not is_available(k,remainder)):
                continue

            values = int_as_sum(remainder, available_multipliers, r)
            if values is not None:
                from sage.combinat.designs.orthogonal_arrays import wilson_construction
                return (wilson_construction,
                        (None,k,r,m,[[(x,1) for x in values]]))

    return False

from sage.combinat.designs.designs_pyx cimport _OA_cache, _OA_cache_size
cdef int is_available(int k, int n) except -1:
    r"""
    Return whether Sage can build an OA(k,n)

    INPUT:

    - ``k``, ``n`` -- integers
    """
    if n >= _OA_cache_size:
        return orthogonal_array(k,n,existence=True) is True
    if k <= _OA_cache[n].max_true:
        return True
    elif k >= _OA_cache[n].min_unknown:
        return False
    else:
        return orthogonal_array(k,n,existence=True) is True
